# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Pervasive/commonly-used utilities."""

import abc
import datetime as dt
import errno
import hashlib
import math
import re
import six
import socket
import ssl
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from oslo_log import log as logging
from oslo_utils import units
from pyasn1.codec.der import decoder as der_decoder
from pyasn1_modules import rfc2459

from pypowervm import const
from pypowervm.i18n import _

# Set up logging
LOG = logging.getLogger(__name__)

XPATH_DELIM = '/'


def dice_href(href, include_scheme_netloc=False, include_query=True,
              include_fragment=True):
    """Parse, sanitize, and reassemble an href.

    :param href: A full link string of the form
                 '<scheme>://<netloc>/<path>;<params>?<query>#<fragment>'.
                 This method also works if the <scheme>://<netloc> is omitted,
                 (but obviously include_scheme_netloc has no effect).
    :param include_scheme_netloc: If True, the <scheme>://<netloc> portion is
                                  included in the returned string.  If False,
                                  it is stripped.
    :param include_query: If True, any ?<query> portion of the link will be
                          included in the return value.
    :param include_fragment: If True, any #<fragment> portion of the link will
                             be included in the return value.
    :return: A string representing the specified portion of the input link.
    """
    parsed = urlparse.urlparse(href)
    ret = ''
    if include_scheme_netloc:
        ret += parsed.scheme + '://' + parsed.netloc
    ret += parsed.path
    # trim trailing '/'s from path, if present
    while ret.endswith('/'):
        ret = ret[:-1]
    if include_query and parsed.query:
        ret += '?' + parsed.query
    if include_fragment and parsed.fragment:
        ret += '#' + parsed.fragment
    return ret


def check_and_apply_xag(path, xag):
    """Validate extended attribute groups and produce the correct path.

    If the existing path already has a group=* other than None, we use it.
    However, if there is a proposed xag - including [] - it must match the
    existing xag, or ValueError is raised.

    Otherwise, we construct the group=* query param according to the
    proposed xag list, as follows:

    If xag is None, use group=None.
    If xag is [] (the empty list), omit the group= query param entirely.
    Otherwise the group= value is a sorted, comma-separated string of the
    xag list.  E.g. for xag=['b', 'c', 'a'], produce 'group=a,b,c'.

    :param path: Input path or href, which may or may not contain a query
                 string, which may or may not contain a group=*.  (Multiple
                 group=* not handled.)  Values in the group=* must be
                 alpha sorted.
    :param xag: Iterable of proposed extended attribute values to be included
                in the query string of the resulting path.
    :return: path, with at most one group=* in the query string.  That
             group= query param's value will be alpha sorted.
    """
    parsed = urlparse.urlsplit(path)
    # parse_qs returns { 'key': ['value'], ... }
    qparms = urlparse.parse_qs(parsed.query) if parsed.query else {}
    path_xag = qparms.pop('group', ['None'])[0]
    if xag is None:
        arg_xag = 'None'
    else:
        # Ensure we have a mutable copy to sort
        xag = list(xag)
        xag.sort()
        arg_xag = ','.join(map(str, xag))  # may be ''

    if path_xag == 'None':
        # No existing xag.  (Treat existing 'group=None' as if not there.)
        # Use whatever was proposed (which may be implicit group=None or
        # may be nothing).
        path_xag = arg_xag
    elif arg_xag != 'None':
        # There was xag in the path already, as well as proposed xag (maybe
        # empty).  Previous xag must match proposed xag if specified
        # (including empty).
        if path_xag != arg_xag:
            raise ValueError(_("Proposed extended attribute group "
                               "'%(arg_xag)s' doesn't match existing "
                               "extended attribute group '%(path_xag)s'") %
                             dict(arg_xag=arg_xag, path_xag=path_xag))
    # else proposed xag is None, so use whatever was already in the path,

    # Whatever we decided on, add it back to the query params if nonempty.
    if path_xag != '':
        qparms['group'] = [path_xag]

    # Rebuild the querystring.  Honor multiples (e.g. foo=bar&foo=baz).
    # (We didn't expect/handle multiple group=*, but need to support it in
    # other keys.)
    qstr = '&'.join(['%s=%s' % (key, val)
                     for key, vals in qparms.items()
                     for val in vals])
    return urlparse.urlunsplit((parsed.scheme, parsed.netloc, parsed.path,
                                qstr, parsed.fragment))


def extend_basepath(href, add):
    """Extends the base path of an href, accounting for querystring/fragment.

    For example, extend_basepath('http://server:1234/foo?a=b&c=d#frag', '/bar')
    => 'http://server:1234/foo/bar?a=b&c=d#frag'

    :param href: Path or href to augment.  Scheme, netloc, query string, and
                 fragment are allowed but not required.
    :param add: String to add onto the base path of the href.  Must not contain
                unescaped special characters such as '?', '&', '#'.
    :return: The augmented href.
    """
    parsed = urlparse.urlsplit(href)
    basepath = parsed.path + add
    return urlparse.urlunsplit((parsed.scheme, parsed.netloc, basepath,
                                parsed.query, parsed.fragment))


def is_instance_path(href):
    """Does the path or href represent an instance (end with UUID)?

    :param href: Path or href to check.  Scheme, netloc, query string, and
                 fragment are allowed but not required.
    :return: True if href's path ends with a UUID, indicating that it
             represents an instance (as opposed to a Feed or some special URI
             such as quick or search).
    """
    path = dice_href(href, include_scheme_netloc=False, include_query=False,
                     include_fragment=False)

    return re.match(const.UUID_REGEX_WORD, path.rsplit('/', 1)[1])


# TODO(IBM): fix (for MITM attacks) or remove (if using loopback only)
def validate_certificate(host, port, certpath, certext):
    hostname = re.sub('[:.]', '_', host)
    cert_file = '%s%s%s' % (certpath, hostname, certext)
    try:
        with open(cert_file, 'r') as f:
            # Retrieve previously trusted certificate
            trusted_cert = ssl.PEM_cert_to_DER_cert(f.read())
    except Exception:
        # found no trusted certificate
        return False
    # Read current certificate from host
    conn = None
    try:
        # workaround for http://bugs.python.org/issue11811
        # should go back to using get_server_certificate when fixed
        # (Issue is resolved as of python 3.3.  Workaround still needed for
        # python 2.7 support.)
        #   rawcert = ssl.get_server_certificate((host, port))
        #   current_cert = ssl.PEM_cert_to_DER_cert(rawcert)
        conn = socket.create_connection((host, port))
        sock = ssl.wrap_socket(conn)
        current_cert = sock.getpeercert(True)
    except Exception:
        # couldn't get certificate from host
        return False
    finally:
        if conn is not None:
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
    # Verify certificate finger prints are the same
    if not (hashlib.sha1(trusted_cert).digest() ==
            hashlib.sha1(current_cert).digest()):
        return False
    # check certificate expiration
    try:
        cert = der_decoder.decode(current_cert,
                                  asn1Spec=rfc2459.Certificate())[0]
        tbs = cert.getComponentByName('tbsCertificate')
        validity = tbs.getComponentByName('validity')
        not_after = validity.getComponentByName('notAfter').getComponent()
        not_after = dt.datetime.strptime(str(not_after), '%y%m%d%H%M%SZ')
        if dt.datetime.utcnow() >= not_after:
            LOG.warning(_('Certificate has expired.'))
            return False
    except Exception:
        LOG.exception('error parsing cert for expiration check')
        return False
    return True


def get_req_path_uuid(path, preserve_case=False, root=False):
    """Extract request target uuid of sanitized path.

    :param path: Path or URI from which to extract the UUID.
    :param preserve_case: If False, the returned UUID will be lowercased.  If
                          True, it will be returned as it exists in the path.
    :param root: If True, and path represents a CHILD entry, the UUID of the
                 ROOT is returned.  Otherwise, the UUID of the target is
                 returned.
    """
    ret = None
    p = dice_href(path, include_query=False, include_fragment=False)
    if '/' in p:
        for maybe_id in p.rsplit('/', 3)[1::2]:
            uuid_match = re.match(const.UUID_REGEX_WORD, maybe_id)
            if uuid_match:
                ret = maybe_id if preserve_case else maybe_id.lower()
                if root:
                    # Want to return the first one.  (If it's a ROOT path, this
                    # will also happen to be the last one.)
                    break
    return ret


def get_uuid_xag_from_path(path):
    uuid = get_req_path_uuid(path)
    parsed = urlparse.urlsplit(path)
    # parse_qs returns { 'key': ['value'], ... }
    qparms = urlparse.parse_qs(parsed.query) if parsed.query else {}
    return uuid.lower(), qparms.get('group', [None])[0]


def convert_bytes_to_gb(bytes_, low_value=.0001, dp=None):
    """Converts an integer of bytes to a decimal representation of gigabytes.

    If the value is too low, will return the 'low_value'.  This is useful
    for converting a small number of bytes (ex. 50) into gigabytes.  Rounding
    may be required.

    :param bytes_: The integer number of bytes.
    :param low_value: The minimum value that should be returned.  (Note: if dp
                      is also specified, the value returned may be rounded up
                      and thus be higher than low_value.)
    :param dp: If specified, the value is rounded up to the specified number of
               decimal places by round_gb_size_up.  (Note: None and zero are
               very different.)
    :returns: The decimal value.
    """
    gb_size = bytes_ / float(units.Gi)
    if gb_size < low_value:
        gb_size = low_value
    if dp is not None:
        gb_size = round_gb_size_up(gb_size, dp=dp)
    return gb_size


def round_gb_size_up(gb_size, dp=2):
    """Rounds a GB disk size (as a decimal float) up to suit the platform.

    Use this method to ensure that new vdisks, LUs, etc. are big enough, as the
    platform generally rounds inputs to the nearest [whatever].  For example, a
    disk of size 4.321GB may wind up at 4.32GB after rounding, possibly leaving
    insufficient space for the image.
    :param gb_size: A decimal float representing the GB size to be rounded.
    :param dp: The number of decimal places to round (up) to.  May be zero
    (round to next highest integer) or negative, (e.g. -1 will round to the
    next highest ten).
    :return: A new decimal float which is greater than or equal to the input.
    """
    shift = 10.0**dp
    return float(math.ceil(gb_size * shift))/shift


def sanitize_mac_for_api(mac):
    """Converts a generalized mac address to one for the API.

    Takes any standard mac (case-insensitive, with or without colons) and
    formats it to uppercase and removes colons.  This is the format for
    the API.
    :param mac: The input mac.
    :returns: The sanitized mac.
    """
    return mac.replace(':', '').upper()


def sanitize_bool_for_api(bool_val):
    """Sanitizes a boolean value for use in the API."""
    return str(bool_val).lower()


def sanitize_float_for_api(float_val, precision=2):
    """Sanitizes a float value for use in the API."""
    template = '%.' + str(precision) + 'f'
    return template % float(float_val)


def sanitize_percent_for_api(float_val, precision=2):
    """Sanitizes a percent value for use in the API.

    :param float_val: A float where valid values are 0.0 <= x <= 1.0. For
                      example the input 0.02 will produce output '2%'.
    :return: A string representation of the passed percentage.
    """
    percent_float = float(float_val)
    if percent_float < 0 or percent_float > 1:
        raise ValueError('A float value 0 <= x <= 1.0 must be provided.')
    percent_float *= 100
    percent_float = sanitize_float_for_api(percent_float, precision)
    return str(percent_float) + '%'


def sanitize_wwpn_for_api(wwpn):
    """Updates the format of the WWPN to match the expected PowerVM format.

    :param wwpn: The original WWPN.
    :return: A WWPN of the format expected by the API.
    """
    return wwpn.upper().replace(':', '')


def sanitize_file_name_for_api(name, prefix='', suffix='',
                               max_len=const.MaxLen.FILENAME_DEFAULT):
    """Generate a sanitized file name based on PowerVM's FileName.Pattern.

    :param name: The base name to sanitize.
    :param prefix: (Optional) A prefix to prepend to the 'name'.  No delimiter
                   is added.
    :param suffix: (Optional) A suffix to append to the 'name'.  No delimiter
                   is added.
    :param max_len: (Optional) The maximum allowable length of the final
                    sanitized string.  Defaults to the API's defined length for
                    FileName.Pattern.
    :return: A string scrubbed of all forbidden characters and trimmed for
             length as necessary.
    """
    def _scrub(in_name):
        """Returns in_name with illegal characters replaced with '_'."""
        return re.sub(r'[^.0-9A-Z_a-z]', '_', in_name)

    name, prefix, suffix = (_scrub(val) for val in (name, prefix, suffix))
    base_len = max_len - len(prefix) - len(suffix)
    if base_len <= 0:
        raise ValueError(_("Prefix and suffix together may not be more than "
                           "%d characters."), max_len - 1)
    name = name[:base_len]
    ret = prefix + name + suffix
    if not len(ret):
        raise ValueError(_("Total length must be at least 1 character."))
    return ret


def sanitize_partition_name_for_api(name, trunc_ok=True):
    """Sanitize a string to be suitable for use as a partition name.

    PowerVM's partition name restrictions are:
    - Between 1 and 31 characters, inclusive;
    - Containing ASCII characters between 0x20 (space) and 0x7E (~), inclusive,
      except ()\<>*$&?|[]'"`

    :param name: The name to scrub.  Invalid characters will be replaced with
                 '_'.
    :param trunc_ok: If True, and name exceeds 31 characters, it is truncated.
                     If False, and name exceeds 31 characters, ValueError is
                     raised.
    :return: The scrubbed string.
    :raise ValueError: If name is None or zero length; or if it exceeds length
                       31 and trunk_ok=False.
    """
    max_len = 31
    if not name:
        raise ValueError(_("The name parameter must be at least one character "
                           "long."))
    if not trunc_ok and len(name) > max_len:
        raise ValueError(_("The name parameter must not exceed %d characters "
                           "when trunk_ok is False."), max_len)
    return re.sub(r'[^- !#%+,./0-9:;=@A-Z^_a-z{}]', '_', name)[:max_len]


def find_equivalent(elem, find_list):
    """Returns the element from the list that is equal to the one passed in.

    For remove operations and what not, the exact object may need to be
    provided.  This method will find the functionally equivalent element
    from the list.

    :param elem: The original element.
    :param find_list: The list to search through.
    :returns: An element from the that is functionally equivalent (based on
              __eq__).  If it does not exist, None is returned.
    """
    for find_elem in find_list:
        if find_elem == elem:
            return find_elem
    return None


def find_wrapper(haystack, needle_uuid):
    """Finds the corresponding wrapper from a list given the UUID.

    :param haystack:  A list of wrappers.  Usually generated from a 'feed' that
                      has been loaded via the wrapper's wrap(response) method.
    :param needle_uuid: The UUID of the object to find in the list.
    :return: The corresponding wrapper for that UUID.  If not found, None.
    """
    for wrapper in haystack:
        if wrapper.uuid == needle_uuid:
            return wrapper
    return None


def xpath(*toks):
    """Constructs an XPath out of the passed-in string components."""
    return XPATH_DELIM.join(toks)


def part_id_by_loc_code(loc_code):
    """Get a partition short ID for a provided virtual device location code.

    All location codes on a virtual device are of the form:
    <MachineType>.<Model>.<Serial>-V<PartID>-C<SlotNumber>

    :return: An int of the associated partition short ID.
    """
    id_match = re.search('.*-V(.+?)-.*', loc_code)
    return int(id_match.group(1)) if id_match else None


def xag_attrs(xagstr, base=const.DEFAULT_SCHEMA_ATTR):
    """Produce XML attributes for a property using extended attribute groups.

    :param xagstr: Extended attribute group name (from pypowervm.const.XAG).
    :param base: The dict of attributes to which to add the extended attribute
                 group.  Usually one of the pypowervm.const values near
                 DEFAULT_SCHEMA_ATTR (the default).
    :return: Dict of XML attributes suitable for the 'attrib' kwarg of a
             (pypowervm.entities or etree) Element constructor.
    """
    return dict(base, group=xagstr) if xagstr else base


def my_partition_id():
    """Return the short ID (not UUID) of the current partition, as an int."""
    with open('/proc/ppc64/lparcfg') as lparcfg:
        for line in lparcfg:
            if line.startswith('partition_id='):
                return int(line.split('=')[1].rstrip())


def parent_spec(parent, parent_type, parent_uuid):
    """Produce a canonical parent type and UUID suitable for read().

    :param parent: EntryWrapper representing the parent.  If specified,
                   parent_type and parent_uuid are ignored.
    :param parent_type: EntryWrapper class or schema_type string representing
                        the schema type of the parent.
    :param parent_uuid: String UUID of the parent.
    :return parent_type: String schema type of the parent.  The parent_type and
                         parent_uuid returns are both None or both valid
                         strings.
    :return parent_uuid: String UUID of the parent.  The parent_type and
                         parent_uuid returns are both None or both valid
                         strings.
    :raise ValueError: If parent is None and parent_type xor parent_uuid is
                       specified.
    """
    if all(param is None for param in (parent, parent_type, parent_uuid)):
        return None, None
    if parent is not None:
        return parent.schema_type, parent.uuid
    if any(param is None for param in (parent_type, parent_uuid)):
        # parent_type xor parent_uuid specified
        raise ValueError(_("Developer error: partial parent specification."))
    # Allow either string or class for parent_type
    if hasattr(parent_type, 'schema_type'):
        parent_type = parent_type.schema_type
    elif type(parent_type) is not str:
        raise ValueError(_("Developer error: parent_type must be either a "
                           "string schema type or a Wrapper subclass."))
    return parent_type, parent_uuid


def retry_io_command(base_cmd, *argv):
    """PEP475: Retry syscalls if EINTR signal received.

    https://www.python.org/dev/peps/pep-0475/

    Certain system calls can be interrupted by signal 4 (EINTR) for no good
    reason.  Per PEP475, these signals should be ignored.  This is implemented
    by default at the lowest level in py3, but we have to account for it in
    py2.

    :param base_cmd: The syscall to wrap.
    :param argv: Arguments to the syscall.
    :return: The return value from invoking the syscall.
    """
    while True:
        try:
            return base_cmd(*argv)
        except EnvironmentError as enve:
            if enve.errno != errno.EINTR:
                raise


@six.add_metaclass(abc.ABCMeta)
class _AllowedList(object):
    """For REST fields taking 'ALL', 'NONE', or [list of values].

    Subclasses should override parse_val and sanitize_for_api.
    """
    ALL = 'ALL'
    NONE = 'NONE'
    _GOOD_STRINGS = (ALL, NONE)

    @staticmethod
    def parse_val(val):
        """Parse a single list value from string to appropriate native type.

        :param val: A single value to parse.
        :return: The converted value.
        """
        # Default native type: str
        return val

    @staticmethod
    def sanitize_for_api(val):
        """Convert a native value to the expected string format for REST.

        :param val: The native value to convert.
        :return: Sanitized string value suitable for the REST API.
        :raise ValueError: If the string can't be converted.
        """
        # Default: Just string-convert
        return str(val)

    @classmethod
    def unmarshal(cls, rest_val):
        """Convert value from REST to a list of vals or an accepted string."""
        rest_val = rest_val.strip()
        if rest_val in cls._GOOD_STRINGS:
            return rest_val
        return [cls.parse_val(val) for val in rest_val.split()]

    @classmethod
    def const_or_list(cls, val):
        """Return one of the _GOOD_STRINGS, or the (sanitized) original list.

        :param val: One of:
                    - A string representing one of the _GOOD_STRINGS (case-
                      insensitive.
                    - A list containing a single value as above.
                    - A list containing values appropriate to the subclass.
        :return: One of:
                 - A string representing one of the _GOOD_STRINGS (in the
                   appropriate case).
                 - A list of the original values, validated and sanitized for
                   the REST API.
                 The objective is to be able to pass the return value directly
                 into a setter or bld method expecting the relevant type.
        :raise ValueError: If the input could not be interpreted/sanitized as
                           appropriate to the subclass.
        """
        ret = None
        if isinstance(val, str) and val.upper() in cls._GOOD_STRINGS:
            ret = val.upper()
        elif isinstance(val, list):
            if (len(val) == 1 and isinstance(val[0], str)
                    and val[0].upper() in cls._GOOD_STRINGS):
                ret = val[0].upper()
            else:
                ret = [cls.sanitize_for_api(ival) for ival in val]
        if ret is not None:
            return ret
        # Not a list, not a good value
        raise ValueError(_("Invalid value '%(bad_val)s'.  Expected one of "
                           "%(good_vals)s, or a list.") %
                         {'bad_val': val, 'good_vals': str(cls._GOOD_STRINGS)})

    @classmethod
    def marshal(cls, val):
        """Produce a string suitable for the REST API."""
        val = cls.const_or_list(val)
        return (' '.join([str(ival) for ival in val]) if isinstance(val, list)
                else val)


class VLANList(_AllowedList):
    """For REST fields of type AllowedVLANIDs.Union."""
    @staticmethod
    def parse_val(val):
        return int(val)

    @staticmethod
    def sanitize_for_api(val):
        try:
            return int(val)
        except (ValueError, TypeError):
            raise ValueError("Specify a list of VLAN integers or integer "
                             "strings; or 'ALL' for all VLANS or 'NONE' for "
                             "no VLANS.")


class MACList(_AllowedList):
    """For REST fields of type AllowedMACAddresses.Union."""
    # Default parse_val is fine

    @staticmethod
    def sanitize_for_api(val):
        return sanitize_mac_for_api(val)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers, constants, and helpers around ManagementConsole."""

from oslo_log import log as logging

import pypowervm.const as c
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.mtms as mtmwrap

LOG = logging.getLogger(__name__)

_MGD_FRAMES = 'ManagedFrames'
_MGD_SYS = 'ManagedSystems'
_MGMT_CON_NAME = 'ManagementConsoleName'

# NETI XPath constants
_NETI_ROOT = 'NetworkInterfaces'

_MGMT_TPLT_OBJ_MOD_VERS = 'TemplateObjectModelVersion'
_MGMT_USR_OBJ_MOD_VERS = 'UserObjectModelVersion'
_MGMT_VERS_INFO = 'VersionInfo'
_MGMT_LOC_VIOS_IMG_NAMES = 'LocalVirtualIOServerImageNames'
_MGMT_WEB_OBJ_MOD_VERS = 'WebObjectModelVersion'

_PWR_ENT_POOLS = 'PowerEnterprisePools'

# SSH Config
_PUB_KEY = 'PublicSSHKey'
_AUTH_KEYS = 'AuthorizedKeys'
_AUTH_KEY = 'AuthorizedKey'

_MGMT_NETI_ROOT = 'ManagementConsoleNetworkInterface'
_MGMT_NETI_NAME = 'InterfaceName'
_MGMT_NETI_ADDRESS = 'NetworkAddress'

_CONS_EL_ORDER = (
    mtmwrap.MTMS_ROOT, _MGD_FRAMES, _MGD_SYS, _MGMT_CON_NAME, _NETI_ROOT,
    _MGMT_TPLT_OBJ_MOD_VERS, _MGMT_USR_OBJ_MOD_VERS, _MGMT_VERS_INFO,
    _MGMT_LOC_VIOS_IMG_NAMES, _MGMT_WEB_OBJ_MOD_VERS, _PWR_ENT_POOLS, _PUB_KEY,
    _AUTH_KEYS)


@ewrap.EntryWrapper.pvm_type('ManagementConsole', child_order=_CONS_EL_ORDER)
class ManagementConsole(ewrap.EntryWrapper):
    """The PowerVM ManagementConsole.

    This refers to the console that is managing PowerVM system. It's the
    one providing the REST API interface.
    """
    @property
    def name(self):
        return self._get_val_str(_MGMT_CON_NAME)

    @property
    def mtms(self):
        return mtmwrap.MTMS.wrap(self.element.find(mtmwrap.MTMS_ROOT))

    @property
    def network_interfaces(self):
        return NetworkInterfaces.wrap(self.element.find(_NETI_ROOT))

    @property
    def ssh_public_key(self):
        return self._get_val_str(_PUB_KEY)

    @property
    def ssh_authorized_keys(self):
        """Returns a list of keys.

        The returned tuple contains the keys as plain strings.
        """
        return tuple(key_w.key for key_w in
                     ewrap.WrapperElemList(self._find_or_seed(_AUTH_KEYS),
                                           AuthorizedKey))

    @ssh_authorized_keys.setter
    def ssh_authorized_keys(self, keys):
        """Sets the keys given a list of key strings."""
        self.replace_list(
            _AUTH_KEYS, [AuthorizedKey.bld(self.adapter, key) for key in keys],
            attrib=c.ATTR_SCHEMA_KSV130)


@ewrap.ElementWrapper.pvm_type(_AUTH_KEY, attrib={})
class AuthorizedKey(ewrap.ElementWrapper):
    """The Authorized Key wrapper."""

    @classmethod
    def bld(cls, adapter, key):
        new_key = super(AuthorizedKey, cls)._bld(adapter)
        new_key.key = key
        return new_key

    @property
    def key(self):
        return self.element.text

    @key.setter
    def key(self, val):
        self.element.text = val


@ewrap.ElementWrapper.pvm_type(_NETI_ROOT, has_metadata=True)
class NetworkInterfaces(ewrap.ElementWrapper):
    """The Network Interfaces wrapper."""

    @property
    def console_interface(self):
        return ConsoleNetworkInterfaces.wrap(
            self.element.find(_MGMT_NETI_ROOT))


@ewrap.ElementWrapper.pvm_type(_MGMT_NETI_ROOT, has_metadata=True)
class ConsoleNetworkInterfaces(ewrap.ElementWrapper):
    """The Console Network Interfaces wrapper."""

    @property
    def name(self):
        return self._get_val_str(_MGMT_NETI_NAME)

    @property
    def address(self):
        return self._get_val_str(_MGMT_NETI_ADDRESS)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Pervasive, widely-used constants."""

import six

_DEFAULT_SCHEMA_VERSION = 'V1_0'
_SCHEMA_VER120 = 'V1_2_0'
_SCHEMA_VER130 = 'V1_3_0'
_SCHEMA_VER140 = 'V1_4_0'
_SCHEMA_VER150 = 'V1_5_0'
_ATTR_SCHEMA_VER = 'schemaVersion'
_ATTR_KSV = 'ksv'
DEFAULT_SCHEMA_ATTR = {_ATTR_SCHEMA_VER: _DEFAULT_SCHEMA_VERSION}
ATTR_KSV120 = {_ATTR_KSV: _SCHEMA_VER120}
ATTR_KSV130 = {_ATTR_KSV: _SCHEMA_VER130}
ATTR_KSV140 = {_ATTR_KSV: _SCHEMA_VER140}
ATTR_KSV150 = {_ATTR_KSV: _SCHEMA_VER150}
ATTR_SCHEMA_KSV130 = {_ATTR_KSV: _SCHEMA_VER130,
                      _ATTR_SCHEMA_VER: _SCHEMA_VER130}

API_BASE_PATH = '/rest/api/'
LOGON_PATH = API_BASE_PATH + 'web/Logon'
TYPE_TEMPLATE = 'application/vnd.ibm.powervm.%s+xml; type=%s'
# The following is interpolated *twice*.  The first time, we insert either the
# Password element or the GenerateX-API-SessionFile element after the UserID.
# We don't want to interpolate 'userid' until the second interpolation, which
# happens at runtime in the Session's login routine.
_LOGONREQUEST_TEMPLATE_TEMPLATE = six.u(
    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n' +
    '<LogonRequest xmlns="http://www.ibm.com/xmlns/systems/power/firmware/' +
    'web/mc/2012_10/" schemaVersion="V1_0">\n' +
    '    <UserID>%%(userid)s</UserID>\n' +
    '    %(pass_or_file)s\n' +
    '</LogonRequest>')
_PASS_TEMPLATE = '<Password>%(passwd)s</Password>'
_SESS_FILE = '<GenerateX-API-SessionFile>true</GenerateX-API-SessionFile>'
# LogonRequest template to be used for password-based authentication
LOGONREQUEST_TEMPLATE_PASS = _LOGONREQUEST_TEMPLATE_TEMPLATE % dict(
    pass_or_file=_PASS_TEMPLATE)
# LogonRequest template to be used for file-based authentication
LOGONREQUEST_TEMPLATE_FILE = _LOGONREQUEST_TEMPLATE_TEMPLATE % dict(
    pass_or_file=_SESS_FILE)

ATOM_NS = 'http://www.w3.org/2005/Atom'
XSI_NS = 'http://www.w3.org/2001/XMLSchema-instance'
WEB_NS = 'http://www.ibm.com/xmlns/systems/power/firmware/web/mc/2012_10/'
PCM_NS = 'http://www.ibm.com/xmlns/systems/power/firmware/pcm/mc/2012_10/'
UOM_BASE_NS = 'http://www.ibm.com/xmlns/systems/power/firmware/uom/mc'
UOM_NS = UOM_BASE_NS + '/2012_10/'

# Match a UUID anywhere in the search string
UUID_REGEX = '%(x)s{8}-%(x)s{4}-%(x)s{4}-%(x)s{4}-%(x)s{12}' % {
    'x': '[A-Fa-f0-9]'}
# Entire search string must be a UUID and nothing more
UUID_REGEX_WORD = '^%s$' % UUID_REGEX
# XPath to the UUID of a metadata-having XML object
UUID_XPATH = 'Metadata/Atom/AtomID'

SUFFIX_TYPE_DO = 'do'
LINK = 'link'

PORT_DEFAULT_BY_PROTO = {
    'http': 12080,
    'https': 12443
}

SERVICE_BY_NS = {
    WEB_NS: 'web',
    UOM_NS: 'uom',
    PCM_NS: 'pcm'
}


class HTTPStatus(object):
    """Small subset of HTTP status codes as used by PowerVM."""
    OK_NO_CONTENT = 204
    NO_CHANGE = 304
    UNAUTHORIZED = 401
    NOT_FOUND = 404
    ETAG_MISMATCH = 412
    INTERNAL_ERROR = 500
    SERVICE_UNAVAILABLE = 503


class MaxLen(object):
    """Maximum lengths for various PowerVM entities."""
    # FileName.Pattern
    FILENAME_DEFAULT = 79
    VOPT_NAME = 37
    VDISK_NAME = 15


class XAG(object):
    """Enumeration of all extended attribute group strings."""
    ALL = 'All'
    NONE = 'None'
    ADV = 'Advanced'
    ENERGY = 'EnergyManagement'
    HYP = 'Hypervisor'
    NVRAM = 'NVRAM'
    SYS_NET = 'SystemNetwork'
    TIER_THRESH = 'TierThreshold'
    VIO_FMAP = 'ViosFCMapping'
    VIO_NET = 'ViosNetwork'
    VIO_SMAP = 'ViosSCSIMapping'
    VIO_STOR = 'ViosStorage'
    VNIC = 'VNIC'



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for properties of EntryWrapper/ElementWrapper subclasses."""

import testtools

import pypowervm.const as c
from pypowervm.utils import wrappers as wutil
from pypowervm.wrappers import base_partition as bp
from pypowervm.wrappers import enterprise_pool as epool
from pypowervm.wrappers import entry_wrapper as ewrap
from pypowervm.wrappers import logical_partition as lpar
from pypowervm.wrappers import virtual_io_server as vios


class TestXAGs(testtools.TestCase):

    def verify_xags(self, wcls, expected_xags):
        """Verify extended attribute groups for properties of a wrapper class.

        :param wcls: The pypowervm.wrappers.entry_wrapper.Wrapper subclass to
                     test.
        :param expected_xags: A dict mapping wcls's property names to their
                              respective extended attribute group names.  Can
                              (should) only include those properties for which
                              an extended attribute group is registered.  (If
                              it contains any other properties, the value must
                              be None.)  Format is { prop_name: xag_name }
        """
        for prop in dir(wcls):
            actual = wcls.get_xag_for_prop(prop)
            expected = expected_xags.get(prop, None)
            self.assertEqual(expected, actual,
                             message="%s.%s" % (wcls.__name__, prop))

    def test_xags(self):
        """Verify xags associated with properties of wrapper classes."""
        # The following wrapper classes have no properties with xags
        for wcls in wutil.wrapper_class_iter():
            if wcls is vios.VIOS:
                self.verify_xags(wcls, {
                    'media_repository': c.XAG.VIO_STOR,
                    'ip_addresses': c.XAG.VIO_NET,
                    'vfc_mappings': c.XAG.VIO_FMAP,
                    'scsi_mappings': c.XAG.VIO_SMAP,
                    'seas': c.XAG.VIO_NET,
                    'trunk_adapters': c.XAG.VIO_NET,
                    'phys_vols': c.XAG.VIO_STOR,
                    'io_adpts_for_link_agg': c.XAG.VIO_NET,
                    'nvram': c.XAG.NVRAM,
                })
            elif wcls is epool.Pool:
                self.verify_xags(wcls, {
                    'compliance_hours_left': c.XAG.ADV
                })
            elif wcls is epool.PoolMember:
                self.verify_xags(wcls, {
                    'proc_compliance_hours_left': c.XAG.ADV,
                    'mem_compliance_hours_left': c.XAG.ADV
                })
            elif wcls in (bp.BasePartition, lpar.LPAR):
                self.verify_xags(wcls, {
                    'nvram': c.XAG.NVRAM
                })
            # Include an elif for each Wrapper subclass that has xags defined.
            else:
                self.verify_xags(wcls, {})

    def test_wrapper_registration(self):
        """All wrapper subclasses must be registered via [base_]pvm_type."""
        for wcls in wutil.wrapper_class_iter():
            if wcls not in (ewrap.Wrapper, ewrap.ElementWrapper,
                            ewrap.EntryWrapper):
                self.assertTrue(wcls._registered,
                                "%s not registered" % wcls.__name__)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import unittest

from pypowervm.i18n import _

# Ensure we're using our test message catalog
os.environ['PYPOWERVM_LOCALEDIR'] = (
    os.path.join(os.path.dirname(os.path.realpath(__file__)), 'locale'))
# Ensure we're using the expected language
os.environ['LANG'] = 'en_US'


class TranslationTests(unittest.TestCase):
    """Test internationalization library."""

    def test_translate(self):
        self.assertEqual(_("This is a test"), "This is an English test")

        self.assertEqual(
            _("This is a message for which a translation doesn't exist"),
            "This is a message for which a translation doesn't exist")



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import pypowervm.const as pc
import pypowervm.entities as ent
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.cluster as clust
import pypowervm.wrappers.mtms as mtmwrap
import pypowervm.wrappers.storage as stor

CLUSTER_RESP = ('<uom:Cluster xmlns:uom='
                '"http://www.ibm.com/xmlns/systems/power/'
                'firmware/uom/mc/2012_10/"'
                ' schemaVersion="V1_0"><uom:Metadata><uom:Atom/>'
                '</uom:Metadata><uom:ClusterName>neotest</uom:ClusterName>'
                '<uom:RepositoryDisk schemaVersion="V1_0"><uom:PhysicalVolume '
                'schemaVersion="V1_0"><uom:Metadata><uom:Atom/></uom:Metadata>'
                '<uom:VolumeName>hdisk123</uom:VolumeName>'
                '</uom:PhysicalVolume></uom:RepositoryDisk><uom:Node '
                'schemaVersion="V1_0"><uom:Node schemaVersion="V1_0">'
                '<uom:Metadata><uom:Atom/></uom:Metadata><uom:HostName>'
                'a.example.com</uom:HostName></uom:Node></uom:Node>'
                '</uom:Cluster>')


class TestCluster(twrap.TestWrapper):

    file = 'cluster.txt'
    wrapper_class_to_test = clust.Cluster

    def test_name(self):
        self.assertEqual(self.dwrap.name, 'neoclust1')

    def test_id(self):
        self.assertEqual(self.dwrap.id, '22cfc907d2abf511e4b2d540f2e95daf30')

    def test_ssp_uri(self):
        self.assertEqual(self.dwrap.ssp_uri, 'https://9.1.2.3:12443/rest/api'
                         '/uom/SharedStoragePool/e357a79a-7a3d-35b6-8405-55ab'
                         '6a2d0de7')

    def test_ssp_uuid(self):
        self.assertEqual(self.dwrap.ssp_uuid.lower(),
                         'e357a79a-7a3d-35b6-8405-55ab6a2d0de7')

    def test_repos_pv(self):
        repos = self.dwrap.repos_pv
        # PV is tested elsewhere.  Minimal verification here.
        self.assertEqual(repos.name, 'hdisk2')
        # Test setter
        newrepos = stor.PV.wrap(
            ent.Element(
                "PhysicalVolume", None,
                attrib={'schemaVersion': 'V1_2_0'},
                children=[
                    ent.Element('Metadata', None,
                                children=[ent.Element('Atom', None)]),
                    ent.Element('VolumeName', None, text='hdisk99')]))
        self.dwrap.repos_pv = newrepos
        self.assertEqual(self.dwrap.repos_pv.name, 'hdisk99')
        # Now try the same thing, but using factory constructor to build PV
        newrepos = stor.PV.bld(None, name='hdisk123')
        self.dwrap.repos_pv = newrepos
        self.assertAlmostEqual(self.dwrap.repos_pv.name, 'hdisk123')

    def test_nodes(self):
        """Tests the Node and MTMS wrappers as well."""
        nodes = self.dwrap.nodes
        self.assertEqual(3, len(nodes))
        node = nodes[0]
        self.assertEqual(node.hostname, 'foo.example.com')
        self.assertEqual(node.lpar_id, 2)
        self.assertEqual(
            node.vios_uri, 'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
            '98498bed-c78a-3a4f-b90a-4b715418fcb6/VirtualIOServer/58C9EB1D-'
            '7213-4956-A011-77D43CC4ACCC')
        self.assertEqual(
            node.vios_uuid, '58C9EB1D-7213-4956-A011-77D43CC4ACCC')
        self.assertEqual(clust.NodeState.UP, nodes[0].state)
        # Validate other NodeState enum values
        self.assertEqual(clust.NodeState.DOWN, nodes[1].state)
        self.assertEqual(clust.NodeState.UNKNOWN, nodes[2].state)
        # Make sure the different Node entries are there
        self.assertEqual(nodes[1].hostname, 'bar.example.com')
        # Test MTMS
        mtms = node.mtms
        self.assertEqual(mtms.machine_type, '8247')
        self.assertEqual(mtms.model, '22L')
        self.assertEqual(mtms.serial, '2125D1A')
        # Test nodes setters
        node2 = nodes[1]
        nodes.remove(node)
        self.assertEqual(2, len(self.dwrap.nodes))
        node._hostname('blah.example.com')
        node._lpar_id(9)
        node._vios_uri('https://foo')
        self.dwrap.nodes = [node2, node]
        self.assertEqual(len(self.dwrap.nodes), 2)
        node = self.dwrap.nodes[1]
        self.assertEqual(node.hostname, 'blah.example.com')
        self.assertEqual(node.lpar_id, 9)
        self.assertEqual(node.vios_uri, 'https://foo')
        # MTMS needs a little more depth
        node._mtms('1234-567*ABCDEF0')
        mtms = node.mtms
        self.assertEqual(mtms.machine_type, '1234')
        self.assertEqual(mtms.model, '567')
        self.assertEqual(mtms.serial, 'ABCDEF0')
        # Now try with a MTMS ElementWrapper
        node._mtms(mtmwrap.MTMS.bld(None, '4321-765*0FEDCBA'))
        mtms = node.mtms
        self.assertEqual(mtms.machine_type, '4321')
        self.assertEqual(mtms.model, '765')
        self.assertEqual(mtms.serial, '0FEDCBA')

    def test_wrapper_classes(self):
        # Cluster
        self.assertEqual(clust.Cluster.schema_type, 'Cluster')
        self.assertEqual(clust.Cluster.schema_ns, pc.UOM_NS)
        self.assertTrue(clust.Cluster.has_metadata)
        self.assertEqual(clust.Cluster.default_attrib, pc.DEFAULT_SCHEMA_ATTR)
        # Node
        self.assertEqual(clust.Node.schema_type, 'Node')
        self.assertEqual(clust.Node.schema_ns, pc.UOM_NS)
        self.assertTrue(clust.Node.has_metadata)
        self.assertEqual(clust.Node.default_attrib, pc.DEFAULT_SCHEMA_ATTR)

    def test_bld_cluster(self):
        n1 = clust.Node.bld(None, hostname='a.example.com')
        repos = stor.PV.bld(None, name='hdisk123')
        cl = clust.Cluster.bld(None, 'foo', repos, n1)
        self.assertEqual(cl.name, 'foo')
        self.assertEqual(cl.repos_pv.name, 'hdisk123')
        self.assertEqual(cl.schema_type, 'Cluster')
        self.assertEqual(cl.schema_ns, pc.UOM_NS)
        nodes = cl.nodes
        self.assertEqual(len(nodes), 1)
        node = nodes[0]
        self.assertEqual(node.hostname, 'a.example.com')
        self.assertEqual(node.schema_type, clust.Node.schema_type)
        self.assertEqual(node.schema_ns, pc.UOM_NS)
        # Node.bld()
        n2 = clust.Node.bld(None, hostname='b.example.com', lpar_id=2,
                            mtms='ABCD-XYZ*1234567', vios_uri='https://foo')
        nodes.append(n2)
        self.assertEqual(len(nodes), 2)
        node = nodes[1]
        self.assertEqual(node.hostname, 'b.example.com')
        self.assertEqual(node.lpar_id, 2)
        self.assertEqual(node.mtms.mtms_str, 'ABCD-XYZ*1234567')
        self.assertEqual(node.vios_uri, 'https://foo')
        self.assertEqual(node.schema_type, clust.Node.schema_type)
        self.assertEqual(node.schema_ns, pc.UOM_NS)

    def test_cluster_ordering(self):
        node = clust.Node.bld(None, hostname='a.example.com')
        repos = stor.PV.bld(None, name='hdisk123')
        cl = clust.Cluster.bld(None, 'neotest', repos, node)
        self.assertEqual(cl.toxmlstring(), CLUSTER_RESP.encode('utf-8'))

if __name__ == "__main__":
    unittest.main()



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Utilities to map slot numbers to I/O elements and vice versa.

These utilities facilitate rebuilding the storage/network topology of an LPAR
e.g. on the target system of a remote rebuild.
"""

import collections
import copy
import pickle
import six
import warnings

from pypowervm import exceptions as pvm_ex
from pypowervm.i18n import _
from pypowervm import util as pvm_util
from pypowervm.utils import lpar_builder as lb
from pypowervm.wrappers import iocard as ioc
from pypowervm.wrappers import managed_system as sys
from pypowervm.wrappers import network as net
from pypowervm.wrappers import storage as stor


class IOCLASS(object):
    """Enumeration of differently-handled I/O classes."""
    VFC = 'VFC'
    LU = stor.LU.__name__
    VDISK = stor.VDisk.__name__
    VOPT = stor.VOptMedia.__name__
    PV = stor.PV.__name__
    CNA = net.CNA.__name__
    MGMT_CNA = 'MGMT' + net.CNA.__name__
    VNIC = ioc.VNIC.__name__


class SlotMapStore(object):
    """Save/fetch slot-to-I/O topology for an LPAR.

    This class should be extended by something that can interact with a
    persistent storage device to implement the save, load, and delete methods.

    The slot metadata is used during a rebuild operation (e.g. Remote Restart)
    to ensure that the client devices are in the same slots on the target.
    Typically this map is constructed on the source system and then saved.  It
    is loaded on the target system and used to rebuild an LPARthe same way.
    """

    def __init__(self, inst_key, load=True):
        """Load (or create) a SlotMapStore for a given LPAR.

        :param inst_key: Unique key (e.g. LPAR UUID) by which the slot map for
                         a given LPAR is referenced in storage.
        :param load: If True (the default), the load method is invoked to
                     retrieve existing data from backing storage.  If False
                     (e.g. if the caller knows there's nothing in the backing
                     store for inst_key; or deliberately wants to replace it),
                     this instance is initialized with an empty topology map.
        """
        self.inst_key = inst_key
        self._vswitch_map = None
        map_str = self.load() if load else None
        # Deserialize or initialize
        self._slot_topo = pickle.loads(map_str) if map_str else {}
        # Save a copy of the topology so we can tell when it has changed
        self._loaded_topo = copy.deepcopy(self._slot_topo)

    @property
    def serialized(self):
        """Internal use only.  Do not override.  Do not invoke."""
        # Used by the save method to serialize the slot map data to an opaque
        # value to write to external storage.
        # Use py2/3-compatible protocol.
        return pickle.dumps(self.topology, protocol=2)

    def load(self):
        """Internal use only.  Do not override.  Do not invoke."""
        # Used by __init__ to bootstrap this instance with saved data, if any.
        return self._load(self.inst_key)

    def _load(self, key):
        """Subclass implementation to load a slot map for an LPAR from storage.

        The subclass must implement this method to retrieve the slot map - an
        opaque data blob - from a storage back-end, where it was stored keyed
        on 'key'.

        :param key: Unique key to associate with this instance in the storage
                    back-end.
        :return: Opaque data blob loaded from storage.  If no value exists in
                 storage for this key, this method should return None.
        """
        return None

    def save(self):
        """Save this slot map to storage, if needed."""
        # Only save if needed.
        if self._slot_topo != self._loaded_topo:
            self._save(self.inst_key, self.serialized)
            # Update the saved version of the loaded topology so a subsequent
            # save (without intervening changes) can condition properly.
            self._loaded_topo = copy.deepcopy(self._slot_topo)

    def _save(self, key, blob):
        """Subclass implementation to write this slot map to storage.

        The subclass must implement this method to save 'blob' - an opaque data
        blob - to a storage back-end, keyed via 'key'.  The object must be
        retrievable subsequently via the same key.  If the back-end already
        contains a value for that key, this method must overwrite it.

        This method need not implement save-only-if-changed logic; that is
        handled by the 'save' method.

        :param key: Unique key to associate with this instance in the storage
                    back-end.
        :param blob: Opaque data blob to save to the storage back-end.  This
                     value is not to be introspected, as its format may change
                     without notice.
        """
        pass

    def delete(self):
        """Remove the back-end storage for this slot map."""
        self._delete(self.inst_key)

    def _delete(self, key):
        """Subclass implementation to delete this slot map from storage.

        The subclass must implement this method to remove the opaque data
        string associated with 'key' from the storage back-end.

        :param key: Unique key to associate with this instance in the storage
                    back-end.
        """
        pass

    def register_max_vslots(self, max_vslots):
        """Register the maximum number of virtual slots on the source LPAR.

        :param max_vslots: The maximum number of virtual slots on the LPAR.
        """
        self._slot_topo['_max_vslots'] = max_vslots

    def register_vnet(self, vnet_w):
        """Register the slot number for a CNA or VNIC.

        :param vnet_w: Either a CNA wrapper or a VNIC wrapper.
        :raises: InvalidVirtualNetworkDeviceType: If the wrapper passed in
                 is not a CNA or VNIC this will be raised.
        """
        if isinstance(vnet_w, net.CNA):
            cna_map = self._vswitch_id2name(vnet_w.adapter)
            self._reg_slot(IOCLASS.CNA, vnet_w.mac, vnet_w.slot,
                           extra_spec=cna_map[vnet_w.vswitch_id])
        elif isinstance(vnet_w, ioc.VNIC):
            self._reg_slot(IOCLASS.VNIC, vnet_w.mac, vnet_w.slot)
        else:
            raise pvm_ex.InvalidVirtualNetworkDeviceType(wrapper=vnet_w)

    def drop_vnet(self, vnet_w):
        """Drops the slot number for a CNA or VNIC.

        :param vnet_w: Either a CNA wrapper or a VNIC wrapper.
        :raises: InvalidVirtualNetworkDeviceType: If the wrapper passed in
                 is not a CNA or VNIC this will be raised.
        """
        if isinstance(vnet_w, net.CNA):
            self._drop_slot(IOCLASS.CNA, vnet_w.mac, vnet_w.slot)
        elif isinstance(vnet_w, ioc.VNIC):
            self._drop_slot(IOCLASS.VNIC, vnet_w.mac, vnet_w.slot)
        else:
            raise pvm_ex.InvalidVirtualNetworkDeviceType(wrapper=vnet_w)

    def register_cna(self, cna):
        """Register the slot and switch topology of a client network adapter.

        :deprecated: Use register_vnet instead.
        :param cna: CNA EntryWrapper to register.
        """
        warnings.warn(_("The register_cna method is deprecated! Please use "
                        "the register_vnet method."), DeprecationWarning)
        self.register_vnet(cna)

    def drop_cna(self, cna):
        """Drops the client network adapter from the slot topology.

        :deprecated: Use drop_vnet instead.
        :param cna: CNA EntryWrapper to drop.
        """
        warnings.warn(_("The drop_cna method is deprecated! Please use "
                        "the drop_vnet method."), DeprecationWarning)
        self.drop_vnet(cna)

    def register_vfc_mapping(self, vfcmap, fab):
        """Incorporate the slot topology associated with a VFC mapping.

        :param vfcmap: VFCMapping ElementWrapper representing the mapping to
                       be incorporated.
        :param fab: The fabric name associated with the mapping.
        """
        self._reg_slot(IOCLASS.VFC, fab, vfcmap.server_adapter.lpar_slot_num)

    def drop_vfc_mapping(self, vfcmap, fab):
        """Drops the client network adapter from the slot topology.

        :param vfcmap: VFCMapping ElementWrapper representing the mapping to
                       be removed.
        :param fab: The fabric name associated with the mapping.
        """
        self._drop_slot(IOCLASS.VFC, fab, vfcmap.server_adapter.lpar_slot_num)

    @staticmethod
    def _parse_vscsi_mapping(vscsimap):
        """Splits out a VSCSIMapping object for use in the internal slot map.

        :param vscsimap: A VSCSIMapping ElementWrapper to process.
        :return: The backing storage element, which may be any of the storage
                 types supported by VSCSIMapping.  None if the vscsimap lacks
                 either backing storage or server adapter.
        :return: The stg_key (storage key) appropriate to the backing storage
                 type.  See the topology @property.  None if the vscsimap lacks
                 either backing storage or server adapter.
        :return: The integer client slot number to which the VSCSIMapping is
                 attached.  None if the vscsimap lacks either backing storage
                 or server adapter.
        :return: The extra_spec (extra specification) value appropriate to the
                 backing storage type.  See the topology @property.  None if
                 the vscsimap lacks either backing storage or server adapter.
        """

        # Must have backing storage and a server adapter to register
        if any(attr is None for attr in (vscsimap.backing_storage,
                                         vscsimap.server_adapter)):
            return None, None, None, None

        extra_spec = None
        bstor = vscsimap.backing_storage
        cslot = vscsimap.server_adapter.lpar_slot_num
        stg_key = bstor.udid
        if isinstance(bstor, stor.VOptMedia):
            # Virtual Optical Media - The IDs will be different on the
            # target.  Using the UDID as a key (above) will at least allow us
            # to determine how many VIOSes should have VOptMedia devices.  For
            # remote rebuild, assuming the consumer uses consistent image
            # naming, we can use the extra_spec to identify which of multiple
            # VOpts we should pick up.
            extra_spec = bstor.name
        else:
            # For shared storage (PV/LU) and for local (VDISK), we need to
            # make sure the LUA (Logical Unit Address) of the device is
            # preserved on the target.  This informs things like boot order.
            extra_spec = vscsimap.target_dev.lua

        return bstor, stg_key, cslot, extra_spec

    def register_vscsi_mapping(self, vscsimap):
        """Incorporate the slot topology associated with a VSCSI mapping.

        :param vscsimap: VSCSIMapping ElementWrapper to be incorporated into
                         the slot topology.
        """
        bstor, stg_key, cslot, extra_spec = self._parse_vscsi_mapping(vscsimap)
        if bstor:
            self._reg_slot(bstor.__class__.__name__, stg_key, cslot,
                           extra_spec=extra_spec)

    def drop_vscsi_mapping(self, vscsimap):
        """Drops the vscsi mapping from the slot topology.

        :param vscsimap: VSCSIMapping ElementWrapper to be removed from the
                         slot topology.
        """
        bstor, stg_key, cslot = self._parse_vscsi_mapping(vscsimap)[:3]

        if bstor:
            self._drop_slot(bstor.__class__.__name__, stg_key, cslot)

    @property
    def topology(self):
        """Produce the slot-to-I/O topology structure from this SlotMapStore.

        :return: A dict of the form:

        { slot_num: { IOCLASS: { io_key: extra_spec } } }

        ...where:

        - slot_num: Integer client slot ID.
        - IOCLASS: The IOCLASS enum indicating the type of I/O.  Each IOCLASS
                   enum is only present if the source had at least one I/O
                   element of that type.
        - io_key: The unique identifier of the mapped I/O element.  This
                  differs by IOCLASS type - see below.
        - extra_spec: Additional information about the I/O element.  This
                      differs by IOCLASS type - see below.

        IOCLASS     stg_key                     extra_spec
        ==============================================================
        CNA         CNA.mac                     VSwitch.name
        VOPT        VOptMedia.udid              Media name
        VDISK       VDisk.udid                  VDisk.capacity (float)
        PV          PV.udid                     LUA
        LU          LU.udid                     LUA
        VFC         fabric name                 None
        VNIC        VNIC.mac                    None
        """
        ret = copy.deepcopy(self._slot_topo)
        ret.pop('_max_vslots', None)
        return ret

    @property
    def max_vslots(self):
        """Returns the highest slot number for the LPAR, or None if not set.

        This number corresponds to base_partition.max_virtual_slots, and
        indicates the maximum number of virtual slots (I/O buses) allowed on
        the LPAR.  If register_max_vslots has never bene called, this value
        will be None.
        """
        return self._slot_topo.get('_max_vslots', None)

    def _vswitch_id2name(self, adap):
        """(Cache and) return a map of VSwitch short ID to VSwitch name.

        :param adap: pypowervm.adapter.Adapter through which to query the
                     VSwitch topology.
        :return: Dict of { vswitch_short_id: vswitch_name }
        """
        if self._vswitch_map is None:
            self._vswitch_map = {
                vsw.switch_id: vsw.name for vsw in net.VSwitch.get(
                    adap, parent=sys.System.get(adap)[0])}
        return self._vswitch_map

    def _reg_slot(self, io_class, io_key, client_slot, extra_spec=None):
        """Register a slot ID where an I/O key can be connected to many slots.

        :param io_class: Outer key representing one of the major classes of I/O
                         handled by SlotMapStore.  Must be one of the IOCLASS
                         enum values.
        :param io_key: Unique identifier of the I/O element to be used as the
                       secondary key.  This differs based on io_class - see the
                       topology @property.
        :param client_slot: The integer slot number by which the I/O element is
                            attached to the client.
        :param extra_spec: Optional extra value to associate with the io_key.
                           This should always be the same for a given io_key.
                           The format/meaning of the value differs based on
                           io_class = see the topology @property.
        """
        # See the topology @property
        # { slot_num: { IOCLASS: { io_key: extra_spec } } }
        if client_slot not in self._slot_topo:
            self._slot_topo[client_slot] = {}
        if io_class not in self._slot_topo[client_slot]:
            self._slot_topo[client_slot][io_class] = {}
        # Always overwrite the extra_spec
        self._slot_topo[client_slot][io_class][io_key] = extra_spec

    def _drop_slot(self, io_class, io_key, client_slot):
        """Drops a client slot ID entry from the topology.

        :param io_class: Outer key representing one of the major classes of I/O
                         handled by SlotMapStore.  Must be one of the IOCLASS
                         enum values.
        :param io_key: Unique identifier of the I/O element to be used as the
                       secondary key.  This differs based on io_class - see the
                       topology @property.
        :param client_slot: The integer slot number by which the I/O element is
                            now detached from the client.
        """
        # See the topology @property
        # { slot_num: { IOCLASS: { io_key: extra_spec } } }
        if client_slot not in self._slot_topo:
            return
        if io_class not in self._slot_topo[client_slot]:
            return
        # Remove the key if it is in the topology
        if io_key in self._slot_topo[client_slot][io_class]:
            del self._slot_topo[client_slot][io_class][io_key]
            # Remove empty internal dicts
            if not self._slot_topo[client_slot][io_class]:
                del self._slot_topo[client_slot][io_class]
            if not self._slot_topo[client_slot]:
                del self._slot_topo[client_slot]


class BuildSlotMap(object):
    """Provides information on which slots should be used for LPAR creates.

    This class takes in a SlotMapStore and provides information on which
    slots should be used on the client adapters.

    If not overridden, this base implementation returns a client slot that
    allows the deploy implementation to choose the 'next available' slot.
    """

    def __init__(self, slot_store):
        """Initializes the slot map.

        :param slot_store: The existing instances SlotMapStore.
        """
        self._slot_store = slot_store
        self._build_map = {}

    def get_vscsi_slot(self, vios_w, udid):
        """Gets the vSCSI client slot and extra spec for the VSCSI device.

        :param vios_w: VIOS wrapper.
        :param udid: UDID of the VSCSI device.
        :return: Integer client slot number on which to create the VSCSIMapping
                 from the specified VIOS for the storage with the specified
                 udid.
        :return: Extra specification appropriate to the storage type.  See the
                 SlotMapStore.topology @property.
        """
        # Pull from the build map.  Will default to None (indicating to
        # fuse an existing vscsi mapping or use next available slot for the
        # mapping).
        # Since the UDID should be universally unique, search all storage types
        for by_vuuid in six.itervalues(self._build_map):
            if vios_w.uuid in by_vuuid and udid in by_vuuid[vios_w.uuid]:
                return by_vuuid[vios_w.uuid][udid]
        return None, None

    def get_pv_vscsi_slot(self, vios_w, udid):
        """DEPRECATED; Gets the vSCSI client slot for the PV.

        Use get_vscsi_slot.  This method will be removed shortly.

        :param vios_w: VIOS wrapper.
        :param udid: UDID of the physical volume.
        :return: Integer client slot number on which to create the VSCSIMapping
                 from the specified VIOS for the PV with the specified udid.
        """
        # Pull from the build map.  Will default to None (indicating to
        # fuse an existing vscsi mapping or use next available slot for the
        # mapping).
        pv_vscsi_map = self._build_map.get(IOCLASS.PV, {})
        return pv_vscsi_map.get(vios_w.uuid, {}).get(udid, (None,))[0]

    def get_vea_slot(self, mac):
        """Gets the client slot for the VEA.

        :param mac: MAC address string to look up.
        :return: Integer client slot number on which to create a CNA with the
                 specified MAC address.
        """
        # Pull from the build map.  Will default to None (indicating to use
        # the next available high slot).
        return self._build_map.get(IOCLASS.CNA, {}).get(
            pvm_util.sanitize_mac_for_api(mac), None)

    def get_mgmt_vea_slot(self):
        """Gets the client slot and MAC for the mgmt VEA.

        :return: MAC Address for the NIC.
        :return: Integer client slot number for the NIC.
        """
        mgmt_vea = self._build_map.get(IOCLASS.MGMT_CNA, {})
        return mgmt_vea.get('mac', None), mgmt_vea.get('slot', None)

    def get_vnet_slot(self, mac):
        """Gets the client slot for the VEA or VNIC, mgmt VEA not included.

        :param mac: MAC address string to look up.
        :return: Integer client slot number on which to create a CNA or VNIC
                 with the specified MAC address.
        """
        # Pull from the build map. Will default to None (indicating to use
        # the next available high slot).
        mac = pvm_util.sanitize_mac_for_api(mac)
        return (self._build_map.get(IOCLASS.CNA, {}).get(mac, None) or
                self._build_map.get(IOCLASS.VNIC, {}).get(mac, None))

    def get_vfc_slots(self, fabric, number_of_slots):
        """Gets the client slot list for a given NPIV fabric.

        :param fabric: Fabric name.
        :param number_of_slots: The number of slots for the specified fabric.
        :return: List of integer client slot numbers on which to map the given
                 fabric.
        :raises: InvalidHostForRebuildSlotMismatch : if the target server
                 requires more or less slots than the source server had.  If
                 this is a first deploy (ex. a standard BuildSlotMap) will not
                 matter, and will return an array of None's (indicating to use
                 the next available slots).
        """
        number_of_map_slots = len(self._build_map.get(IOCLASS.VFC,
                                                      {}).get(fabric, []))
        if not number_of_map_slots:
            return [None] * number_of_slots
        if number_of_map_slots == number_of_slots:
            return self._build_map.get(IOCLASS.VFC, {}).get(fabric, None)
        raise pvm_ex.InvalidHostForRebuildSlotMismatch(
            rebuild_slots=number_of_slots,
            original_slots=number_of_map_slots)

    def get_max_vslots(self):
        """Retrieve or derive the maximum number of virtual slots for the LPAR.

        If the source LPAR's maximum number of virtual slots was registered in
        the source slot map, that value is returned.  Otherwise, we attempt to
        derive a reasonable value based on the highest registered slot number.
        If none was registered, the minimum returned value will be the default
        from pypowervm.utils.lpar_builder.DEF_MAX_SLOT.

        It is the caller's responsibility to determine whether the value
        returned by this method is sufficiently high for the LPAR being
        created.
        """
        # Return the exact value if it was registered
        if self._slot_store.max_vslots is not None:
            return self._slot_store.max_vslots
        # Otherwise, return the higher of
        # - The default from lpar_builder
        # - The highest registered slot, plus 10
        from_hi_slot = (max(self._slot_store.topology.keys())
                        if self._slot_store.topology else 0) + 10
        return max(lb.DEF_MAX_SLOT, from_hi_slot)


class RebuildSlotMap(BuildSlotMap):
    """Used to determine the slot topology when rebuilding a VM.

    A LPAR rebuild needs to configure the client slots with the exact topology
    as their source.  This implementation requires additional details from the
    target server, but then provides the LPAR's appropriate client slot
    layout.
    """

    def __init__(self, slot_store, vios_wraps, vscsi_vol_to_vio,
                 npiv_fabrics):
        """Initializes the rebuild map.

        :param slot_store: The existing instances SlotMapStore.
        :param vios_wraps: List of VIOS EntryWrappers.  Must have been
                           retrieved with the appropriate XAGs.
        :param vscsi_vol_to_vio: The volume to virtual I/O server mapping.
                                 Of the following format:
                                 { 'lu_udid' : [ 'vios_uuid', 'vios_uuid'],
                                   'pv_udid' : [ 'vios_uuid', 'vios_uuid'] }
        :param npiv_fabrics: List of vFC fabric names.
        """
        super(RebuildSlotMap, self).__init__(slot_store)

        self.vios_wraps = vios_wraps

        # Lets first get the VEAs and VNICs built
        self._vea_build_out()
        self._vnic_build_out()

        # Next up is vSCSI
        self._vscsi_build_out(vscsi_vol_to_vio)

        # And finally vFC (npiv)
        self._npiv_build_out(npiv_fabrics)

    def get_mgmt_vea_slot(self):
        """Gets the client slot and MAC for the mgmt VEA.

        There should only be one MGMT VEA per system.  As such, subsequent
        calls to this will return the single MGMT VEA.

        :return: MAC Address for the NIC.
        :return: Integer client slot number for the NIC.
        """
        # On a rebuild specifically, the MGMT VIF may not be there.  If that
        # is the case, we want to make sure that it is the next available slot.
        mgmt_vea = self._build_map.get(IOCLASS.MGMT_CNA, {})
        slot = mgmt_vea.get('slot')

        # If the slot is None, that means the MGMT vif wasn't there initially
        # but its being requested for the rebuild.
        if slot is None:
            # Make sure there were slots, then use the 'next highest' available
            slots_in_use = sorted(self._slot_store.topology, reverse=True)
            if len(slots_in_use) > 0:
                slot = slots_in_use[0] + 1

        return mgmt_vea.get('mac', None), slot

    def _vscsi_build_slot_order(self):
        """Order slots by (descending) number of storage elements they host.

        :return: An ordered dictionary of the form { slot_num: count } where
                 slot_num is the integer slot number and count is the number of
                 supported* storage elements attached to this slot.  The dict
                 is ordered such that an iterator over its keys will return the
                 slot_num with the highest count first, etc.
                 *Only PV and LU are supported at this time.
        """
        slots_order = {}
        for slot in self._slot_store.topology:
            io_dict = self._slot_store.topology[slot]
            # There are multiple types of things that can go into the vSCSI
            # map.  Some are not supported for rebuild.
            if io_dict.get(IOCLASS.VOPT):
                raise pvm_ex.InvalidHostForRebuildInvalidIOType(
                    io_type='Virtual Optical Media')

            # Create a dictionary of slots to the number of mappings per
            # slot. This will determine which slots we assign first.
            slots_order[slot] = (len(io_dict.get(IOCLASS.PV, {})) +
                                 len(io_dict.get(IOCLASS.LU, {})) +
                                 len(io_dict.get(IOCLASS.VDISK, {})))

        # For VSCSI we need to figure out which slot numbers have the most
        # mappings and assign these ones to VIOSes first in descending order.
        #
        # We do this because if we have 4 mappings for storage elements 1
        # through 4 on slot X and 5 mappings for storage elements 1 through 5
        # on slot Y, then we must ensure the VIOS that has storage elements 1
        # through 5 gets slot Y (even though it's a candidate for slot X). We
        # solve this by assigning the most used slot numbers first.
        slots_order = collections.OrderedDict(sorted(
            six.iteritems(slots_order), key=lambda t: t[1], reverse=True))

        return slots_order

    def _vscsi_build_out(self, vol_to_vio):
        """Builds the '_build_map' for physical volumes and logical units."""
        slots_order = self._vscsi_build_slot_order()

        # We're going to use the vol_to_vio dictionary for consistency and
        # remove elements from it. We need to deepcopy so that the original
        # remains the same.
        vol_to_vio_cp = copy.deepcopy(vol_to_vio)

        for slot in slots_order:
            slot_topo = self._slot_store.topology[slot]
            if not any(slot_topo.get(x) for x in
                       (IOCLASS.PV, IOCLASS.LU, IOCLASS.VDISK)):
                continue
            # Initialize the set of candidate VIOSes to all available VIOSes.
            # We'll filter out and remove any VIOSes that can't host any PV or
            # LU for this slot.
            candidate_vioses = set(vio.uuid for vio in self.vios_wraps)

            slot_info = [(udid, lua, IOCLASS.PV) for udid, lua in
                         six.iteritems(slot_topo.get(IOCLASS.PV, {}))]
            slot_info.extend([(udid, lua, IOCLASS.LU) for udid, lua in
                              six.iteritems(slot_topo.get(IOCLASS.LU, {}))])
            slot_info.extend([(udid, lua, IOCLASS.VDISK) for udid, lua in
                              six.iteritems(slot_topo.get(IOCLASS.VDISK, {}))])
            for udid, lua, stg_class in slot_info:

                # If the UDID isn't anywhere to be found on the destination
                # VIOSes then we have a problem.
                if udid not in vol_to_vio_cp:
                    raise pvm_ex.InvalidHostForRebuildNoVIOSForUDID(udid=udid)

                # Inner Join. The goal is to end up with a set that only has
                # VIOSes which can see every backing storage elem for this
                # slot.
                candidate_vioses &= set(vol_to_vio_cp[udid])

                # If the set of candidate VIOSes is empty then this host is
                # not a candidate for rebuild.
                if not candidate_vioses:
                    raise pvm_ex.InvalidHostForRebuildNotEnoughVIOS(udid=udid)

            # Just take one, doesn't matter which one.
            # TODO(IBM): Perhaps find a way to ensure better distribution.
            vios_uuid_for_slot = candidate_vioses.pop()

            for udid, lua, stg_class in slot_info:

                self._put_vios_val(stg_class, vios_uuid_for_slot, udid,
                                   (slot, lua))

                # There's somewhat of a problem with this. We want to remove
                # the VIOS UUID we're picking from this list so that other
                # VIOSes will pick up the other mappings for this storage, but
                # it may be the case that the original storage actually
                # belonged to more than one mapping on a single VIOS. It's not
                # clear if this is allowed, and if it is the backing storage
                # can potentially be corrupted.
                #
                # If there were multiple mappings for the same vSCSI storage
                # element on the same VIOS then the slot store could not
                # identify it. We may hit an invalid host for rebuild exception
                # if this happens or we may not. It depends on the differences
                # between source and destination VIOSes.
                vol_to_vio_cp[udid].remove(vios_uuid_for_slot)

    def _vea_build_out(self):
        """Builds the '_build_map' for the veas."""
        for slot, io_dict in six.iteritems(self._slot_store.topology):
            for mac, vswitch in six.iteritems(io_dict.get(IOCLASS.CNA, {})):
                mac = pvm_util.sanitize_mac_for_api(mac)
                if vswitch == 'MGMTSWITCH':
                    self._put_mgmt_vea_slot(mac, slot)
                else:
                    self._put_novios_val(IOCLASS.CNA, mac, slot)

    def _vnic_build_out(self):
        """Builds the '_build_map' for the vnics."""
        for slot, io_dict in six.iteritems(self._slot_store.topology):
            for mac in io_dict.get(IOCLASS.VNIC, {}):
                self._put_novios_val(
                    IOCLASS.VNIC, pvm_util.sanitize_mac_for_api(mac), slot)

    def _npiv_build_out(self, fabrics):
        """Builds the build map for the NPIV fabrics.

        :param fabrics: List of NPIV fabric names.
        :raise InvalidHostForRebuildFabricsNotFound: If any fabrics in the
                                                     slot_map topology are not
                                                     in fabrics.
        """
        seen_fabrics = set()
        for fabric in fabrics:
            fabric_slots = []
            # Add the slot numbers for this fabric
            for slot, iomap in six.iteritems(self._slot_store.topology):
                if fabric not in iomap.get(IOCLASS.VFC, {}):
                    continue
                fabric_slots.append(slot)
                seen_fabrics.add(fabric)

            self._put_novios_val(IOCLASS.VFC, fabric, fabric_slots)

        # Make sure all the topology's fabrics are accounted for.
        # topo_fabrics is all the fabrics in all the slots from the slot_map
        # topology.
        topo_fabrics = {fab for iomap in self._slot_store.topology.values()
                        for fab in iomap.get(IOCLASS.VFC, {}).keys()}
        if topo_fabrics - seen_fabrics:
            raise pvm_ex.InvalidHostForRebuildFabricsNotFound(
                fabrics=', '.join(topo_fabrics - seen_fabrics))

    def _put_mgmt_vea_slot(self, mac, slot):
        """Store client slot data for the managament VEA.

        There should only ever be one of these.

        Enhances the rebuild map with:
        { IOCLASS_MGMT_CNA: { 'mac': mac, 'slot': 'slot' } }

        :param mac: MAC address (string) of the management VEA.
        :param slot: Client slot number for the management VEA.
        """
        if IOCLASS.MGMT_CNA not in self._build_map:
            self._build_map[IOCLASS.MGMT_CNA] = {}
        self._build_map[IOCLASS.MGMT_CNA] = {'mac': mac, 'slot': slot}

    def _put_novios_val(self, io_class, io_key, val):
        """Store a keyed value not associated with a VIOS.

        This applies to non-management CNAs, VNICs, and NPIV fabrics.
        Enhances the rebuild map with:
        { io_class: { io_key: val } }

        :param io_class: IOCLASS const value representing the type of I/O.
                         Either IOCLASS.CNA, IOCLASS.VNIC, or IOCLASS.NPIV
        :param io_key: Key of the I/O device to be added.  MAC address for
                       IOCLASS.CNA, IOCLASS.VNIC; fabric name for IOCLASS.VFC.
        :param val: The slot value(s) to be added.  A list of slot numbers for
                    IOCLASS.VFC; a single slot number for IOCLASS.CNA and
                    IOCLASS.VNIC.
        """
        if io_class not in self._build_map:
            self._build_map[io_class] = {}
        self._build_map[io_class][io_key] = val

    def _put_vios_val(self, stg_class, vios_uuid, udid, val):
        """Store client slot info associated with a storage dev and VIOS.

        This applies to VSCSI devices.  Enhances the rebuild map with:
        { stg_class: { vios_uuid: { stg_key: val } } }

        :param stg_class: IOCLASS const value representing the type of storage.
        :param vios_uuid: UUID of the VIOS which will host the storage device
                          indicated by stg_key.
        :param udid: UDID of the storage device to be added.
        :param val: The slot data to be added.  For IOCLASS.PV, this is a tuple
                    of (slot, lua).
        """
        if stg_class not in self._build_map:
            self._build_map[stg_class] = {}
        if vios_uuid not in self._build_map[stg_class]:
            self._build_map[stg_class][vios_uuid] = {}
        self._build_map[stg_class][vios_uuid][udid] = val



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes, enums, and constants shared by LPAR and VIOS EntryWrappers."""
from pypowervm import const
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.iocard as card

# Base Partition (_BP)
_BP_ALLOW_PERF_DATA_COLL = 'AllowPerformanceDataCollection'
_BP_ASSOC_PROF = 'AssociatedPartitionProfile'
_BP_AVAIL_PRIORITY = 'AvailabilityPriority'
_BP_CURR_BSR_ARRAYS = 'CurrentAllocatedBarrierSynchronizationRegisterArrays'
_BP_CURRENT_PROC_MODE = 'CurrentProcessorCompatibilityMode'
_BP_PROFILE_SYNC = 'CurrentProfileSync'
_BP_HOSTNAME = 'Hostname'
_BP_BOOTABLE = 'IsBootable'
_BP_CALL_HOME = 'IsCallHomeEnabled'
_BP_CONN_MONITORING = 'IsConnectionMonitoringEnabled'
_BP_OP_IN_PROGRESS = 'IsOperationInProgress'
_BP_REDUNDANT_ERR_PATH_REP = 'IsRedundantErrorPathReportingEnabled'
_BP_TIME_REF = 'IsTimeReferencePartition'
_BP_ATTN_LED = 'IsVirtualServiceAttentionLEDOn'
_BP_TRUSTED_PLATFORM = 'IsVirtualTrustedPlatformModuleEnabled'
_BP_KEYLOCK_POS = 'KeylockPosition'
_BP_LOGICAL_SERIAL_NUM = 'LogicalSerialNumber'
_BP_OS_VER = 'OperatingSystemVersion'
_BP_CAPABILITIES = 'PartitionCapabilities'
_BP_ID = 'PartitionID'
_BP_IO_CFG = 'PartitionIOConfiguration'
_BP_MEM_CFG = 'PartitionMemoryConfiguration'
_BP_NAME = 'PartitionName'
_BP_PROC_CFG = 'PartitionProcessorConfiguration'
_BP_PROFS = 'PartitionProfiles'
_BP_STATE = 'PartitionState'
_BP_TYPE = 'PartitionType'
_BP_UUID = 'PartitionUUID'
_BP_PENDING_PROC_MODE = 'PendingProcessorCompatibilityMode'
_BP_PROC_POOL = 'ProcessorPool'
_BP_PROG_DATA_REMAIN = 'ProgressPartitionDataRemaining'
_BP_PROG_DATA_TOTAL = 'ProgressPartitionDataTotal'
_BP_PROG_STATE = 'ProgressState'
_BP_RMC_STATE = 'ResourceMonitoringControlState'
_BP_RMC_IP = 'ResourceMonitoringIPAddress'
_BP_VAL_INT_PERF = 'ValidInteractivePerformance'
_BP_ASSOC_SYSTEM = 'AssociatedManagedSystem'
_BP_SRIOV_ETH = 'SRIOVEthernetLogicalPorts'
_BP_SRIOV_ROCE = 'SRIOVRoCELogicalPorts'
_BP_SRIOV_FC_ETH = 'SRIOVFibreChannelOverEthernetLogicalPorts'
_BP_CNAS = 'ClientNetworkAdapters'
_BP_HOST_ETH = 'HostEthernetAdapterLogicalPorts'
_BP_MAC_PREF = 'MACAddressPrefix'
_BP_SVC_PARTITION = 'IsServicePartition'
_BP_MGMT_CAP = 'PowerVMManagementCapable'
_BP_REF_CODE = 'ReferenceCode'
_BP_REF_CODE_FULL = 'ReferenceCodeFull'
_BP_MGT_PARTITION = 'IsManagementPartition'
_BP_AUTO_START = 'AutoStart'
_BP_BOOT_MODE = 'BootMode'
_BP_NVRAM = 'PartitionNVRAM'
_BP_UPTIME = 'Uptime'
_BP_DISABLE_SECURE_BOOT = 'DisableSecureBoot'
_BP_ASSOC_GROUPS = 'AssociatedGroups'
_BP_POWER_ON_WITH_HYP = 'PowerOnWithHypervisor'
_BP_ASSOC_TASKS = 'AssociatedTasks'
_BP_DESC = 'Description'

BP_EL_ORDER = (
    _BP_ALLOW_PERF_DATA_COLL, _BP_ASSOC_PROF, _BP_AVAIL_PRIORITY,
    _BP_CURR_BSR_ARRAYS, _BP_CURRENT_PROC_MODE, _BP_PROFILE_SYNC, _BP_HOSTNAME,
    _BP_BOOTABLE, _BP_CALL_HOME, _BP_CONN_MONITORING, _BP_OP_IN_PROGRESS,
    _BP_REDUNDANT_ERR_PATH_REP, _BP_TIME_REF, _BP_ATTN_LED,
    _BP_TRUSTED_PLATFORM, _BP_KEYLOCK_POS, _BP_LOGICAL_SERIAL_NUM, _BP_OS_VER,
    _BP_CAPABILITIES, _BP_ID, _BP_IO_CFG, _BP_MEM_CFG, _BP_NAME, _BP_PROC_CFG,
    _BP_PROFS, _BP_STATE, _BP_TYPE, _BP_UUID, _BP_PENDING_PROC_MODE,
    _BP_PROC_POOL, _BP_PROG_DATA_REMAIN, _BP_PROG_DATA_TOTAL, _BP_PROG_STATE,
    _BP_RMC_STATE, _BP_RMC_IP, _BP_VAL_INT_PERF, _BP_ASSOC_SYSTEM,
    _BP_SRIOV_ETH, _BP_SRIOV_ROCE, _BP_SRIOV_FC_ETH, _BP_CNAS, _BP_HOST_ETH,
    _BP_MAC_PREF, _BP_SVC_PARTITION, _BP_MGMT_CAP, _BP_REF_CODE,
    _BP_REF_CODE_FULL, _BP_MGT_PARTITION, _BP_AUTO_START, _BP_BOOT_MODE,
    _BP_NVRAM, _BP_UPTIME, _BP_DISABLE_SECURE_BOOT, _BP_ASSOC_GROUPS,
    _BP_POWER_ON_WITH_HYP, _BP_ASSOC_TASKS, _BP_DESC
)

# Partition Capabilities (_CAP)
_CAP_DLPAR_IO_CAPABLE = 'DynamicLogicalPartitionIOCapable'
_CAP_DLPAR_MEM_CAPABLE = 'DynamicLogicalPartitionMemoryCapable'
_CAP_DLPAR_PROC_CAPABLE = 'DynamicLogicalPartitionProcessorCapable'
_CAP_INTRUSION_DETECT_CAPABLE = 'InternalAndExternalIntrusionDetectionCapable'
_CAP_RMC_OS_SHUTDOWN_CAPABLE = ('ResourceMonitoringControlOperatingSystem'
                                'ShutdownCapable')
_CAP_EL_ORDER = (_CAP_DLPAR_IO_CAPABLE, _CAP_DLPAR_MEM_CAPABLE,
                 _CAP_DLPAR_PROC_CAPABLE, _CAP_INTRUSION_DETECT_CAPABLE,
                 _CAP_RMC_OS_SHUTDOWN_CAPABLE,)

# Processor Configuration (_PC)
_PC_DED_PROC_CFG = 'DedicatedProcessorConfiguration'
_PC_HAS_DED_PROCS = 'HasDedicatedProcessors'
_PC_SHR_PROC_CFG = 'SharedProcessorConfiguration'
_PC_SHARING_MODE = 'SharingMode'
_PC_CURR_HAS_DED_PROCS = 'CurrentHasDedicatedProcessors'
_PC_CURR_SHARING_MODE = 'CurrentSharingMode'
_PC_CURR_DED_PROC_CFG = 'CurrentDedicatedProcessorConfiguration'
_PC_RUN_HAS_DED_PROCS = 'RuntimeHasDedicatedProcessors'
_PC_RUN_SHARING_MODE = 'RuntimeSharingMode'
_PC_CURR_SHR_PROC_CFG = 'CurrentSharedProcessorConfiguration'
_PC_EL_ORDER = (_PC_DED_PROC_CFG, _PC_HAS_DED_PROCS, _PC_SHR_PROC_CFG,
                _PC_SHARING_MODE, _PC_CURR_HAS_DED_PROCS,
                _PC_CURR_SHARING_MODE, _PC_CURR_DED_PROC_CFG,
                _PC_RUN_HAS_DED_PROCS, _PC_RUN_SHARING_MODE,
                _PC_CURR_SHR_PROC_CFG)

# Shared Processor Configuration (_SPC)
_SPC_DES_PROC_UNIT = 'DesiredProcessingUnits'
_SPC_DES_VIRT_PROC = 'DesiredVirtualProcessors'
_SPC_MAX_PROC_UNIT = 'MaximumProcessingUnits'
_SPC_MAX_VIRT_PROC = 'MaximumVirtualProcessors'
_SPC_MIN_PROC_UNIT = 'MinimumProcessingUnits'
_SPC_MIN_VIRT_PROC = 'MinimumVirtualProcessors'
_SPC_SHARED_PROC_POOL_ID = 'SharedProcessorPoolID'
_SPC_UNCAPPED_WEIGHT = 'UncappedWeight'
_SPC_ALLOC_VIRT_PROC = 'AllocatedVirtualProcessors'
_SPC_CURR_MAX_PROC_UNIT = 'CurrentMaximumProcessingUnits'
_SPC_CURR_MIN_PROC_UNIT = 'CurrentMinimumProcessingUnits'
_SPC_CURR_PROC_UNIT = 'CurrentProcessingUnits'
_SPC_CURR_SHARED_PROC_POOL_ID = 'CurrentSharedProcessorPoolID'
_SPC_CURR_UNCAPPED_WEIGHT = 'CurrentUncappedWeight'
_SPC_CURR_MIN_VIRT_PROC = 'CurrentMinimumVirtualProcessors'
_SPC_CURR_MAX_VIRT_PROC = 'CurrentMaximumVirtualProcessors'
_SPC_RUN_PROC_UNIT = 'RuntimeProcessingUnits'
_SPC_RUN_UNCAPPED_WEIGHT = 'RuntimeUncappedWeight'
_SPC_EL_ORDER = (_SPC_DES_PROC_UNIT, _SPC_DES_VIRT_PROC, _SPC_MAX_PROC_UNIT,
                 _SPC_MAX_VIRT_PROC, _SPC_MIN_PROC_UNIT, _SPC_MIN_VIRT_PROC,
                 _SPC_SHARED_PROC_POOL_ID, _SPC_UNCAPPED_WEIGHT,
                 _SPC_ALLOC_VIRT_PROC, _SPC_CURR_MAX_PROC_UNIT,
                 _SPC_CURR_MIN_PROC_UNIT, _SPC_CURR_PROC_UNIT,
                 _SPC_CURR_SHARED_PROC_POOL_ID, _SPC_CURR_UNCAPPED_WEIGHT,
                 _SPC_CURR_MIN_VIRT_PROC, _SPC_CURR_MAX_VIRT_PROC,
                 _SPC_RUN_PROC_UNIT, _SPC_RUN_UNCAPPED_WEIGHT)

# Dedicated Processor Configuration (_DPC)
_DPC_DES_PROCS = 'DesiredProcessors'
_DPC_MAX_PROCS = 'MaximumProcessors'
_DPC_MIN_PROCS = 'MinimumProcessors'

# Partition Memory Configuration (_MEM)
_MEM_PROF_AME_ENABLED = 'ActiveMemoryExpansionEnabled'
_MEM_AMS_ENABLED = 'ActiveMemorySharingEnabled'
_MEM_BSR_ARRAY_CT = 'BarrierSynchronizationRegisterArrayCount'
_MEM_DES_ENT = 'DesiredEntitledMemory'
_MEM_DES_HUGE_PAGE_CT = 'DesiredHugePageCount'
_MEM_DES = 'DesiredMemory'
_MEM_EXP_FACTOR = 'ExpansionFactor'
_MEM_HW_PG_TBL_RATIO = 'HardwarePageTableRatio'
_MEM_MAN_ENT_MODE_ENABLED = 'ManualEntitledModeEnabled'
_MEM_MAX_HUGE_PG_CT = 'MaximumHugePageCount'
_MEM_MAX = 'MaximumMemory'
_MEM_WT = 'MemoryWeight'
_MEM_MIN_HUGE_PG_CT = 'MinimumHugePageCount'
_MEM_MIN = 'MinimumMemory'
_MEM_PRI_PGING_SVC_PART = 'PrimaryPagingServicePartition'
_MEM_SEC_PGING_SVC_PART = 'SecondaryPagingServicePartition'
_MEM_AUTO_ENT_MEM_ENABLED = 'AutoEntitledMemoryEnabled'
_MEM_CURR_BSR_ARRAYS = 'CurrentBarrierSynchronizationRegisterArrays'
_MEM_CURR_ENT = 'CurrentEntitledMemory'
_MEM_CURR_EXP_FACT = 'CurrentExpansionFactor'
_MEM_CURR_HW_PG_TBL_RATIO = 'CurrentHardwarePageTableRatio'
_MEM_CURR_HUGE_PG_CT = 'CurrentHugePageCount'
_MEM_CURR_MAX_HUGE_PG_CT = 'CurrentMaximumHugePageCount'
_MEM_CURR_MAX = 'CurrentMaximumMemory'
_MEM_CUR = 'CurrentMemory'
_MEM_CURR_MEM_WT = 'CurrentMemoryWeight'
_MEM_CURR_MIN_HUGE_PG_CT = 'CurrentMinimumHugePageCount'
_MEM_CURR_MIN = 'CurrentMinimumMemory'
_MEM_CURR_PGING_SVC_PART = 'CurrentPagingServicePartition'
_MEM_EXP_HW_ACC_ENABLED = 'MemoryExpansionHardwareAccessEnabled'
_MEM_ENC_HW_ACC_ENABLED = 'MemoryEncryptionHardwareAccessEnabled'
_MEM_AME_ENABLED = 'MemoryExpansionEnabled'
_MEM_RELEASABLE = 'MemoryReleaseable'
_MEM_TO_RELEASE = 'MemoryToRelease'
_MEM_RED_ERR_PATH_REP_ENABLED = 'RedundantErrorPathReportingEnabled'
_MEM_REQ_MIN_FOR_MAX = 'RequiredMinimumForMaximum'
_MEM_RUNT_ENT = 'RuntimeEntitledMemory'
_MEM_RUNT_EXP_FACT = 'RuntimeExpansionFactor'
_MEM_RUNT_HUGE_PG_CT = 'RuntimeHugePageCount'
_MEM_RUNT = 'RuntimeMemory'
_MEM_RUNT_WT = 'RuntimeMemoryWeight'
_MEM_RUNT_MIN = 'RuntimeMinimumMemory'
_MEM_SHARED_MEM_ENABLED = 'SharedMemoryEnabled'

_MEM_EL_ORDER = (
    _MEM_PROF_AME_ENABLED, _MEM_AMS_ENABLED, _MEM_BSR_ARRAY_CT, _MEM_DES_ENT,
    _MEM_DES_HUGE_PAGE_CT, _MEM_DES, _MEM_EXP_FACTOR, _MEM_HW_PG_TBL_RATIO,
    _MEM_MAN_ENT_MODE_ENABLED, _MEM_MAX_HUGE_PG_CT, _MEM_MAX, _MEM_WT,
    _MEM_MIN_HUGE_PG_CT, _MEM_MIN, _MEM_PRI_PGING_SVC_PART,
    _MEM_SEC_PGING_SVC_PART, _MEM_AUTO_ENT_MEM_ENABLED, _MEM_CURR_BSR_ARRAYS,
    _MEM_CURR_ENT, _MEM_CURR_EXP_FACT, _MEM_CURR_HW_PG_TBL_RATIO,
    _MEM_CURR_HUGE_PG_CT, _MEM_CURR_MAX_HUGE_PG_CT, _MEM_CURR_MAX, _MEM_CUR,
    _MEM_CURR_MEM_WT, _MEM_CURR_MIN_HUGE_PG_CT, _MEM_CURR_MIN,
    _MEM_CURR_PGING_SVC_PART, _MEM_EXP_HW_ACC_ENABLED, _MEM_ENC_HW_ACC_ENABLED,
    _MEM_AME_ENABLED, _MEM_RELEASABLE, _MEM_TO_RELEASE,
    _MEM_RED_ERR_PATH_REP_ENABLED, _MEM_REQ_MIN_FOR_MAX, _MEM_RUNT_ENT,
    _MEM_RUNT_EXP_FACT, _MEM_RUNT_HUGE_PG_CT, _MEM_RUNT, _MEM_RUNT_WT,
    _MEM_RUNT_MIN, _MEM_SHARED_MEM_ENABLED)

# Partition I/O Configuration (_IO)
IO_CFG_ROOT = _BP_IO_CFG
_IO_MAX_SLOTS = 'MaximumVirtualIOSlots'
_IO_TIO = 'TaggedIO'

# Tagged I/O (_TIO)
_TIO_ALT_CONSOLE = 'AlternateConsole'
_TIO_ALT_LOAD_SRC = 'AlternateLoadSource'
_TIO_CONSOLE = 'Console'
_TIO_LOAD_SRC = 'LoadSource'
_TIO_OP_CONSOLE = 'OperationsConsole'
_TIO_EL_ORDER = (_TIO_ALT_CONSOLE, _TIO_ALT_LOAD_SRC, _TIO_CONSOLE,
                 _TIO_LOAD_SRC, _TIO_OP_CONSOLE)

# Constants for the I/O Slot Configuration
IO_SLOTS_ROOT = 'ProfileIOSlots'
IO_SLOT_ROOT = 'ProfileIOSlot'
_IO_SLOT_REQ = 'IsRequired'

# Constants for the Associated I/O Slot
ASSOC_IO_SLOT_ROOT = 'AssociatedIOSlot'
_ASSOC_IO_SLOT_BUS_GRP = 'BusGroupingRequired'
_ASSOC_IO_SLOT_DESC = 'Description'
_ASSOC_IO_SLOT_FEAT_CODES = 'FeatureCodes'
_ASSOC_IO_SLOT_PHYS_LOC = 'IOUnitPhysicalLocation'
_ASSOC_IO_SLOT_ADPT_ID = 'PCAdapterID'
_ASSOC_IO_SLOT_PCI_CLASS = 'PCIClass'
_ASSOC_IO_SLOT_PCI_DEV_ID = 'PCIDeviceID'
_ASSOC_IO_SLOT_PCI_SUBSYS_DEV_ID = 'PCISubsystemDeviceID'
_ASSOC_IO_SLOT_PCI_MFG_ID = 'PCIManufacturerID'
_ASSOC_IO_SLOT_PCI_REV_ID = 'PCIRevisionID'
_ASSOC_IO_SLOT_PCI_VENDOR_ID = 'PCIVendorID'
_ASSOC_IO_SLOT_SUBSYS_VENDOR_ID = 'PCISubsystemVendorID'
_ASSOC_IO_SLOT_DRC_INDEX = 'SlotDynamicReconfigurationConnectorIndex'
_ASSOC_IO_SLOT_DRC_NAME = 'SlotDynamicReconfigurationConnectorName'

# Constants for generic I/O Adapter
IO_ADPT_ROOT = 'IOAdapter'
RELATED_IO_ADPT_ROOT = 'RelatedIOAdapter'

_IO_SLOT_ORDER = (ASSOC_IO_SLOT_ROOT, _IO_SLOT_REQ)

_AIO_ORDER = (_ASSOC_IO_SLOT_BUS_GRP, _ASSOC_IO_SLOT_DESC,
              _ASSOC_IO_SLOT_FEAT_CODES, _ASSOC_IO_SLOT_PHYS_LOC,
              _ASSOC_IO_SLOT_ADPT_ID, _ASSOC_IO_SLOT_PCI_CLASS,
              _ASSOC_IO_SLOT_PCI_DEV_ID, _ASSOC_IO_SLOT_PCI_SUBSYS_DEV_ID,
              _ASSOC_IO_SLOT_PCI_MFG_ID, _ASSOC_IO_SLOT_PCI_REV_ID,
              _ASSOC_IO_SLOT_PCI_VENDOR_ID, _ASSOC_IO_SLOT_SUBSYS_VENDOR_ID,
              RELATED_IO_ADPT_ROOT, _ASSOC_IO_SLOT_DRC_INDEX,
              _ASSOC_IO_SLOT_DRC_NAME)

IO_PFC_ADPT_ROOT = 'PhysicalFibreChannelAdapter'
_IO_ADPT_ID = 'AdapterID'
_IO_ADPT_DESC = 'Description'
_IO_ADPT_DEV_NAME = 'DeviceName'
_IO_ADPT_DEV_TYPE = 'DeviceType'
_IO_ADPT_DYN_NAME = 'DynamicReconfigurationConnectorName'
_IO_ADPT_PHYS_LOC = 'PhysicalLocation'
_IO_ADPT_UDID = 'UniqueDeviceID'

PFC_PORT_WWPN = card.PFC_PORT_WWPN
PFC_PORTS_ROOT = card.PFC_PORTS_ROOT
PFC_PORT_ROOT = card.PFC_PORT_ROOT

IOAdapter = card.IOAdapter
PhysFCAdapter = card.PhysFCAdapter
PhysFCPort = card.PhysFCPort


class SharingMode(object):
    """Shared Processor sharing modes.

    Subset of LogicalPartitionProcessorSharingModeEnum.
    """
    CAPPED = 'capped'
    UNCAPPED = 'uncapped'
    ALL_VALUES = (CAPPED, UNCAPPED)


class DedicatedSharingMode(object):
    """Dedicated Processor sharing modes.

    Subset of LogicalPartitionProcessorSharingModeEnum.
    """
    SHARE_IDLE_PROCS = 'sre idle proces'
    SHARE_IDLE_PROCS_ACTIVE = 'sre idle procs active'
    SHARE_IDLE_PROCS_ALWAYS = 'sre idle procs always'
    KEEP_IDLE_PROCS = 'keep idle procs'
    ALL_VALUES = (SHARE_IDLE_PROCS, SHARE_IDLE_PROCS_ACTIVE,
                  SHARE_IDLE_PROCS_ALWAYS, KEEP_IDLE_PROCS)


class LPARState(object):
    """State of a given LPAR.

    From LogicalPartitionStateEnum.
    """
    ERROR = 'error'
    NOT_ACTIVATED = 'not activated'
    NOT_AVAILBLE = 'not available'
    OPEN_FIRMWARE = 'open firmware'
    RUNNING = 'running'
    SHUTTING_DOWN = 'shutting down'
    STARTING = 'starting'
    MIGRATING_NOT_ACTIVE = 'migrating not active'
    MIGRATING_RUNNING = 'migrating running'
    HARDWARE_DISCOVERY = 'hardware discovery'
    SUSPENDED = 'suspended'
    SUSPENDING = 'suspending'
    RESUMING = 'resuming'
    UNKNOWN = 'Unknown'


class LPARType(object):
    """Subset of LogicalPartitionEnvironmentEnum."""
    OS400 = 'OS400'
    AIXLINUX = 'AIX/Linux'
    VIOS = 'Virtual IO Server'


class LPARCompat(object):
    """LPAR compatibility modes.

    From LogicalPartitionProcessorCompatibilityModeEnum.
    """
    DEFAULT = 'default'
    POWER6 = 'POWER6'
    POWER6_PLUS = 'POWER6_Plus'
    POWER6_PLUS_ENHANCED = 'POWER6_Plus_Enhanced'
    POWER7 = 'POWER7'
    POWER8 = 'POWER8'
    POWER9 = 'POWER9'
    ALL_VALUES = (DEFAULT, POWER6, POWER6_PLUS, POWER6_PLUS_ENHANCED, POWER7,
                  POWER8, POWER9)


class RMCState(object):
    """Various RMC States.

    From ResourceMonitoringControlStateEnum.
    """
    ACTIVE = 'active'
    INACTIVE = 'inactive'
    NONE = 'none'
    UNKNOWN = 'unknown'
    BUSY = 'busy'


class BootMode(object):
    """Mirror of PartitionBootMode.Enum.

    Valid values for LPAR.bootmode/VIOS.bootmode.

    Not to be confused with pypowervm.tasks.power.BootMode.

    Example usage:
        lwrap.bootmode = BootMode.NORM
        lwrap.update()
    """
    NORM = 'Normal'
    SMS = 'System_Management_Services'
    DD = 'Diagnostic_With_Default_Boot_List'
    DS = 'Diagnostic_With_Stored_Boot_List'
    OF = 'Open_Firmware'
    UNAVAILABLE = 'Unavailable'
    DEFAULT = 'Default'
    UNKNOWN = 'Unknown'
    ALL_VALUES = (NORM, SMS, DD, DS, OF, UNAVAILABLE, DEFAULT, UNKNOWN)


class KeylockPos(object):
    """Mirror of KeylockPosition.Enum.

    Valid values for LPAR.keylock_pos/VIOS.keylock_pos.

    Not to be confused with pypowervm.tasks.power.KeylockPos.

    Example usage:
        lwrap.keylock_pos = KeylockPos.MANUAL
        lwrap.update()
    """
    MANUAL = 'manual'
    NORMAL = 'normal'
    UNKNOWN = 'unknown'
    ALL_VALUES = (MANUAL, NORMAL, UNKNOWN)


class _DlparCapable(object):
    def _can_modify(self, dlpar_cap, cap_desc):
        """Checks to determine if the partition can be modified.

        :param dlpar_cap: The appropriate DLPAR attribute to validate.  Only
                          used if system is active.
        :param cap_desc: A translated string indicating the DLPAR capability.
        :return capable: True if HW can be added/removed.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of modification.  If capable is True, the
                        reason will be None.
        """
        # First check is the not activated state
        if self.state == LPARState.NOT_ACTIVATED:
            return True, None
        if self.rmc_state != RMCState.ACTIVE and not self.is_mgmt_partition:
            return False, _('Partition does not have an active RMC '
                            'connection.')
        if not dlpar_cap:
            return False, _('Partition does not have an active DLPAR '
                            'capability for %s.') % cap_desc
        return True, None

    def can_modify_io(self):
        """Determines if a partition is capable of adding/removing I/O HW.

        :return capable: True if HW can be added/removed.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of modification.  If capable is True, the
                        reason will be None.
        """
        return self._can_modify(self.capabilities.io_dlpar, _('I/O'))

    def can_modify_mem(self):
        """Determines if a partition is capable of adding/removing Memory.

        :return capable: True if memory can be added/removed.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of modification.  If capable is True, the
                        reason will be None.
        """
        return self._can_modify(self.capabilities.mem_dlpar, _('Memory'))

    def can_modify_proc(self):
        """Determines if a partition is capable of adding/removing processors.

        :return capable: True if procs can be added/removed.  False otherwise.
        :return reason: A translated message that will indicate why it was not
                        capable of modification.  If capable is True, the
                        reason will be None.
        """
        return self._can_modify(self.capabilities.proc_dlpar, _('Processors'))


@ewrap.Wrapper.base_pvm_type
class BasePartition(ewrap.EntryWrapper, _DlparCapable):
    """Base class for Logical Partition (LPAR) & Virtual I/O Server (VIOS).

    This corresponds to the abstract BasePartition object in the PowerVM
    schema.
    """

    search_keys = dict(name=_BP_NAME, id=_BP_ID)

    @classmethod
    def _bld_base(cls, adapter, name, mem_cfg, proc_cfg, env, io_cfg=None):
        """Creates a BasePartition wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the partition
        :param mem_cfg: The memory configuration wrapper
        :param proc_cfg: The processor configuration wrapper
        :param env: The type of partition, taken from LPARType
        :param io_cfg: The I/O configuration wrapper

        :returns: New BasePartition wrapper
        """

        partition = super(BasePartition, cls)._bld(adapter)
        if io_cfg:
            partition.io_config = io_cfg
        partition.mem_config = mem_cfg
        partition.name = name
        partition.proc_config = proc_cfg
        partition._env(env)

        return partition

    @property
    def state(self):
        """See LPARState.

        e.g. 'not activated', 'running', 'migrating running', etc.
        """
        return self._get_val_str(_BP_STATE)

    @property
    def name(self):
        """Short name (not ID, MTMS, or hostname)."""
        return self._get_val_str(_BP_NAME)

    @name.setter
    def name(self, val):
        self.set_parm_value(_BP_NAME, val)

    @property
    def id(self):
        """Short ID (not UUID)."""
        return self._get_val_int(_BP_ID)

    def _id(self, value):
        """Set ID (not UUID). Only settable on creation of the partition."""
        self.set_parm_value(_BP_ID, int(value))

    @property
    def env(self):
        """See the LPARType Enumeration.

        Should usually be 'AIX/Linux' for LPAR.  'Virtual IO Server' should
        only happen for VIOS.
        """
        return self._get_val_str(_BP_TYPE)

    def _env(self, val):
        self.set_parm_value(_BP_TYPE, val)

    @property
    def partition_uuid(self):
        return self._get_val_str(_BP_UUID)

    @property
    def assoc_sys_uuid(self):
        """UUID of the associated ManagedSystem."""
        href = self.get_href(_BP_ASSOC_SYSTEM, one_result=True)
        return u.get_req_path_uuid(href, preserve_case=True) if href else None

    @property
    def rmc_state(self):
        """See RMCState.

        e.g. 'active', 'inactive', 'busy', etc.
        """
        return self._get_val_str(_BP_RMC_STATE)

    @property
    def rmc_ip(self):
        """IP address used for RMC communication, as a string."""
        return self._get_val_str(_BP_RMC_IP)

    @property
    def operating_system(self):
        """String representing the OS and version, or 'Unknown'."""
        return self._get_val_str(_BP_OS_VER, 'Unknown')

    @property
    def ref_code(self):
        return self._get_val_str(_BP_REF_CODE)

    @property
    def ref_code_full(self):
        return self._get_val_str(_BP_REF_CODE_FULL)

    @property
    def avail_priority(self):
        return self._get_val_int(_BP_AVAIL_PRIORITY, 0)

    @avail_priority.setter
    def avail_priority(self, value):
        self.set_parm_value(_BP_AVAIL_PRIORITY, value)

    @property
    def profile_sync(self):
        return self._get_val_str(_BP_PROFILE_SYNC, default='Off') == 'On'

    @profile_sync.setter
    def profile_sync(self, value):
        if type(value) == bool:
            value = 'On' if value else 'Off'
        self.set_parm_value(_BP_PROFILE_SYNC, value)

    @property
    def proc_compat_mode(self):
        """*Current* processor compatibility mode.

        See LPARCompat.  E.g. 'POWER7',
        'POWER7_Plus', 'POWER8', etc.
        """
        return self._get_val_str(_BP_CURRENT_PROC_MODE)

    @property
    def pending_proc_compat_mode(self):
        """Pending processor compatibility mode.

        See LPARCompat.  E.g. 'POWER7',
        'POWER7_Plus', 'POWER8', etc.
        """
        return self._get_val_str(_BP_PENDING_PROC_MODE)

    @proc_compat_mode.setter
    def proc_compat_mode(self, value):
        """Sets *PENDING* proc compat mode.

        Note that corresponding getter retrieves the *CURRENT* proc compat
        mode.
        """
        self.set_parm_value(_BP_PENDING_PROC_MODE, value)

    @property
    def is_mgmt_partition(self):
        """Is this the management partition?  Default False if field absent."""
        return self._get_val_bool(_BP_MGT_PARTITION)

    @property
    def is_service_partition(self):
        """Is this the service partition?  Default False if field absent."""
        return self._get_val_bool(_BP_SVC_PARTITION)

    @is_service_partition.setter
    def is_service_partition(self, value):
        """Set if this is the service partition."""
        self.set_parm_value(_BP_SVC_PARTITION, u.sanitize_bool_for_api(value))

    @property
    def keylock_pos(self):
        """Keylock position - see KeylockPos enumeration."""
        return self._get_val_str(_BP_KEYLOCK_POS)

    @keylock_pos.setter
    def keylock_pos(self, value):
        """Keylock position - see KeylockPos enumeration."""
        if value not in KeylockPos.ALL_VALUES:
            raise ValueError(_("Invalid KeylockPos '%s'.") % value)
        self.set_parm_value(_BP_KEYLOCK_POS, value)

    @property
    def bootmode(self):
        """Boot mode - one of the BootMode enum values."""
        return self._get_val_str(_BP_BOOT_MODE)

    @bootmode.setter
    def bootmode(self, val):
        if val not in BootMode.ALL_VALUES:
            raise ValueError(_("Invalid BootMode '%s'.") % val)
        self.set_parm_value(_BP_BOOT_MODE, val)

    @property
    def disable_secure_boot(self):
        return self._get_val_bool(_BP_DISABLE_SECURE_BOOT)

    @disable_secure_boot.setter
    def disable_secure_boot(self, value):
        self.set_parm_value(
            _BP_DISABLE_SECURE_BOOT, u.sanitize_bool_for_api(value),
            attrib=const.ATTR_KSV150)

    @property
    def allow_perf_data_collection(self):
        return self._get_val_bool(_BP_ALLOW_PERF_DATA_COLL)

    @allow_perf_data_collection.setter
    def allow_perf_data_collection(self, value):
        self.set_parm_value(_BP_ALLOW_PERF_DATA_COLL,
                            u.sanitize_bool_for_api(value))

    @property
    def capabilities(self):
        elem = self._find(_BP_CAPABILITIES)
        return PartitionCapabilities.wrap(elem)

    @property
    def io_config(self):
        """The Partition I/O Configuration."""
        elem = self._find(_BP_IO_CFG)
        return PartitionIOConfiguration.wrap(elem)

    @io_config.setter
    def io_config(self, io_cfg):
        """The Partition I/O Configuration for the LPAR."""
        elem = self._find_or_seed(_BP_IO_CFG)
        # TODO(efried): All instances of _find_or_seed + element.replace should
        # probably be inject instead
        self.element.replace(elem, io_cfg.element)

    @property
    def mem_config(self):
        """The Partition Memory Configuration for the LPAR."""
        elem = self._find(_BP_MEM_CFG)
        return PartitionMemoryConfiguration.wrap(elem)

    @mem_config.setter
    def mem_config(self, mem_cfg):
        """The Partition Memory Configuration for the LPAR."""
        elem = self._find_or_seed(_BP_MEM_CFG)
        self.element.replace(elem, mem_cfg.element)

    @property
    def proc_config(self):
        """The Partition Processor Configuration for the LPAR."""
        elem = self._find(_BP_PROC_CFG)
        return PartitionProcessorConfiguration.wrap(elem)

    @proc_config.setter
    def proc_config(self, proc_config):
        """The Partition Processor Configuration for the LPAR."""
        elem = self._find_or_seed(_BP_PROC_CFG)
        self.element.replace(elem, proc_config.element)

    @ewrap.Wrapper.xag_property(const.XAG.NVRAM)
    def nvram(self):
        return self._get_val_str(_BP_NVRAM)

    @nvram.setter
    def nvram(self, nvram):
        self.set_parm_value(_BP_NVRAM, nvram,
                            attrib=u.xag_attrs(const.XAG.NVRAM,
                                               base=const.ATTR_KSV130))

    @property
    def uptime(self):
        """Integer time since partition boot, in seconds."""
        return self._get_val_int(_BP_UPTIME)


@ewrap.ElementWrapper.pvm_type(_BP_CAPABILITIES, has_metadata=True,
                               child_order=_CAP_EL_ORDER)
class PartitionCapabilities(ewrap.ElementWrapper):
    """See LogicalPartitionCapabilities."""
    @property
    def io_dlpar(self):
        return self._get_val_bool(_CAP_DLPAR_IO_CAPABLE)

    @property
    def mem_dlpar(self):
        return self._get_val_bool(_CAP_DLPAR_MEM_CAPABLE)

    @property
    def proc_dlpar(self):
        return self._get_val_bool(_CAP_DLPAR_PROC_CAPABLE)


@ewrap.ElementWrapper.pvm_type(_BP_PROC_CFG, has_metadata=True,
                               child_order=_PC_EL_ORDER)
class PartitionProcessorConfiguration(ewrap.ElementWrapper):
    """Represents the partitions Processor Configuration.

    Comprised of either the shared or dedicated processor config.
    """

    @classmethod
    def bld_shared(cls, adapter, proc_unit, proc,
                   sharing_mode=SharingMode.UNCAPPED, uncapped_weight=128,
                   min_proc_unit=None, max_proc_unit=None, min_proc=None,
                   max_proc=None, proc_pool=0):
        """Builds a Shared Processor configuration wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param proc_unit: Amount of desired proc units (float)
        :param proc: Number of virtual processors (int)
        :param sharing_mode: Sharing mode of the processors (uncapped)
        :param uncapped_weight: Uncapped weight of the processors (0-255)
        :param min_proc_unit: Minimum proc units, default to proc unit value
        :param max_proc_unit: Maximum proc units, default to proc unit value
        :param min_proc: Minimum processors, default to proc value
        :param max_proc: Maximum processors, default to proc value
        :param proc_pool: The shared processor pool for the lpar, defaults to 0
        :returns: Processor Config with shared processors

        """
        proc_cfg = super(PartitionProcessorConfiguration, cls)._bld(adapter)
        proc_cfg._has_dedicated(False)

        sproc = SharedProcessorConfiguration.bld(
            adapter, proc_unit, proc, uncapped_weight=uncapped_weight,
            min_proc_unit=min_proc_unit, max_proc_unit=max_proc_unit,
            min_proc=min_proc, max_proc=max_proc, proc_pool=proc_pool)

        proc_cfg._shared_proc_cfg(sproc)
        proc_cfg.sharing_mode = sharing_mode
        return proc_cfg

    @classmethod
    def bld_dedicated(cls, adapter, proc, min_proc=None, max_proc=None,
                      sharing_mode=DedicatedSharingMode.SHARE_IDLE_PROCS):

        """Builds a Dedicated Processor configuration wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param proc: Number of virtual processors (int)
        :param min_proc: Minimum processors, default to proc value
        :param max_proc: Maximum processors, default to proc value
        :param sharing_mode: Sharing mode of the processors, 'sre idle proces'
        :returns: Processor Config with dedicated processors

        """

        proc_cfg = super(PartitionProcessorConfiguration, cls)._bld(adapter)

        dproc = DedicatedProcessorConfiguration.bld(
            adapter, proc, min_proc=min_proc, max_proc=max_proc)

        proc_cfg._dedicated_proc_cfg(dproc)
        proc_cfg._has_dedicated(True)
        proc_cfg.sharing_mode = sharing_mode
        return proc_cfg

    @property
    def has_dedicated(self):
        """Returns boolean True if dedicated, False if shared or not found."""
        return self._get_val_bool(_PC_HAS_DED_PROCS)

    def _has_dedicated(self, val):
        """Expects 'true' (string) for dedicated or 'false' for shared."""
        self.set_parm_value(_PC_HAS_DED_PROCS, u.sanitize_bool_for_api(val))

    @property
    def sharing_mode(self):
        return self._get_val_str(_PC_SHARING_MODE)

    @sharing_mode.setter
    def sharing_mode(self, value):
        self.set_parm_value(_PC_SHARING_MODE, value)

    @property
    def shared_proc_cfg(self):
        """Returns the Shared Processor Configuration."""
        return SharedProcessorConfiguration.wrap(
            self._find(_PC_SHR_PROC_CFG))

    def _shared_proc_cfg(self, spc):
        elem = self._find_or_seed(_PC_SHR_PROC_CFG)
        self.element.replace(elem, spc.element)

    @property
    def dedicated_proc_cfg(self):
        """Returns the Dedicated Processor Configuration."""
        return DedicatedProcessorConfiguration.wrap(
            self._find(_PC_DED_PROC_CFG))

    def _dedicated_proc_cfg(self, dpc):
        elem = self._find_or_seed(_PC_DED_PROC_CFG)
        self.element.replace(elem, dpc.element)


@ewrap.ElementWrapper.pvm_type(_BP_MEM_CFG, has_metadata=True,
                               child_order=_MEM_EL_ORDER)
class PartitionMemoryConfiguration(ewrap.ElementWrapper):
    """Represents the partitions Memory Configuration."""

    @classmethod
    def bld(cls, adapter, mem, min_mem=None, max_mem=None):
        """Creates the ParitionMemoryConfiguration.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param mem: The amount of memory for the partition in MB
        :param min_mem: The minimum amount of memory in MB. Defaults to
            the mem param
        :param max_mem: The maximum amount of memory in MB. Defaults to
            the mem param
        :returns: The memory configuration wrapper.
        """
        if min_mem is None:
            min_mem = mem
        if max_mem is None:
            max_mem = mem

        cfg = super(PartitionMemoryConfiguration, cls)._bld(adapter)
        cfg.desired = mem
        cfg.max = max_mem
        cfg.min = min_mem

        return cfg

    @property
    def current(self):
        return self._get_val_int(_MEM_CUR)

    @property
    def desired(self):
        return self._get_val_int(_MEM_DES)

    @desired.setter
    def desired(self, mem):
        self.set_parm_value(_MEM_DES, str(mem))

    @property
    def max(self):
        return self._get_val_int(_MEM_MAX)

    @max.setter
    def max(self, mem):
        self.set_parm_value(_MEM_MAX, str(mem))

    @property
    def min(self):
        return self._get_val_int(_MEM_MIN)

    @min.setter
    def min(self, mem):
        self.set_parm_value(_MEM_MIN, str(mem))

    @property
    def shared_enabled(self):
        # The default is None instead of False so that the caller
        # can know if the value is not set
        return self._get_val_bool(_MEM_SHARED_MEM_ENABLED, None)

    @property
    def ame_enabled(self):
        return self._get_val_bool(_MEM_AME_ENABLED)

    @property
    def exp_factor(self):
        """The Active Memory Expansion Factor

        The expansion factor represents the target memory multiplier.
        e.g. An LPAR with EF = 2 which has 4 GB of memory will have a target
        expansion memory of 8 GB.
        """
        return self._get_val_float(_MEM_EXP_FACTOR, default=0)

    @exp_factor.setter
    def exp_factor(self, exp_factor):
        """The Active Memory Expansion Factor

        :param exp_factor: The expansion factor value. Setting this to 0 will
                           turn/keep AME off. The valid values are
                           1.0 <= x <= 10.0 up to 2 decimal places.
        """
        self.set_parm_value(_MEM_EXP_FACTOR,
                            u.sanitize_float_for_api(exp_factor))


@ewrap.ElementWrapper.pvm_type(_PC_SHR_PROC_CFG, has_metadata=True,
                               child_order=_SPC_EL_ORDER)
class SharedProcessorConfiguration(ewrap.ElementWrapper):
    """Represents the partition's Shared Processor Configuration."""

    @classmethod
    def bld(cls, adapter, proc_unit, proc, uncapped_weight=None,
            min_proc_unit=None, max_proc_unit=None,
            min_proc=None, max_proc=None, proc_pool=0):
        """Builds a Shared Processor configuration wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param proc_unit: Amount of desired proc units (float)
        :param proc: Number of virtual processors (int)
        :param uncapped_weight: Uncapped weight of the processors, 0-255
        :param min_proc_unit: Minimum proc units, default to proc unit value
        :param max_proc_unit: Maximum proc units, default to proc unit value
        :param min_proc: Minimum processors, default to proc value
        :param max_proc: Maximum processors, default to proc value
        :param proc_pool: The shared processor pool for the lpar, defaults to 0
        :returns: Processor Config with shared processors

        """

        # Set defaults if not specified
        if min_proc_unit is None:
            min_proc_unit = proc_unit
        if max_proc_unit is None:
            max_proc_unit = proc_unit
        if min_proc is None:
            min_proc = proc
        if max_proc is None:
            max_proc = proc

        sproc = super(SharedProcessorConfiguration, cls)._bld(adapter)

        sproc.desired_units = proc_unit
        sproc.desired_virtual = proc
        sproc.max_units = max_proc_unit
        sproc.max_virtual = max_proc
        sproc.min_units = min_proc_unit
        sproc.min_virtual = min_proc
        sproc.pool_id = proc_pool
        if uncapped_weight is not None:
            sproc.uncapped_weight = uncapped_weight

        return sproc

    @property
    def desired_units(self):
        return self._get_val_float(_SPC_DES_PROC_UNIT)

    @desired_units.setter
    def desired_units(self, val):
        self.set_parm_value(_SPC_DES_PROC_UNIT, u.sanitize_float_for_api(val))

    @property
    def max_units(self):
        return self._get_val_float(_SPC_MAX_PROC_UNIT)

    @max_units.setter
    def max_units(self, val):
        self.set_parm_value(_SPC_MAX_PROC_UNIT, u.sanitize_float_for_api(val))

    @property
    def min_units(self):
        return self._get_val_float(_SPC_MIN_PROC_UNIT)

    @min_units.setter
    def min_units(self, val):
        self.set_parm_value(_SPC_MIN_PROC_UNIT, u.sanitize_float_for_api(val))

    @property
    def desired_virtual(self):
        return self._get_val_int(_SPC_DES_VIRT_PROC)

    @desired_virtual.setter
    def desired_virtual(self, val):
        self.set_parm_value(_SPC_DES_VIRT_PROC, val)

    @property
    def max_virtual(self):
        return self._get_val_int(_SPC_MAX_VIRT_PROC)

    @max_virtual.setter
    def max_virtual(self, val):
        self.set_parm_value(_SPC_MAX_VIRT_PROC, val)

    @property
    def min_virtual(self):
        return self._get_val_int(_SPC_MIN_VIRT_PROC)

    @min_virtual.setter
    def min_virtual(self, val):
        self.set_parm_value(_SPC_MIN_VIRT_PROC, val)

    @property
    def pool_id(self):
        return self._get_val_int(_SPC_SHARED_PROC_POOL_ID, 0)

    @pool_id.setter
    def pool_id(self, val):
        self.set_parm_value(_SPC_SHARED_PROC_POOL_ID, val)

    @property
    def uncapped_weight(self):
        return self._get_val_int(_SPC_UNCAPPED_WEIGHT, 0)

    @uncapped_weight.setter
    def uncapped_weight(self, val):
        self.set_parm_value(_SPC_UNCAPPED_WEIGHT, val)


@ewrap.ElementWrapper.pvm_type(_PC_DED_PROC_CFG, has_metadata=True)
class DedicatedProcessorConfiguration(ewrap.ElementWrapper):
    """Represents the partition's Dedicated Processor Configuration."""

    @classmethod
    def bld(cls, adapter, proc, min_proc=None, max_proc=None):
        """Builds a Dedicated Processor configuration wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param proc: Number of virtual processors (int)
        :param min_proc: Minimum processors, default to proc value
        :param max_proc: Maximum processors, default to proc value
        :returns: Processor Config with dedicated processors

        """

        # Set defaults if not specified
        if min_proc is None:
            min_proc = proc
        if max_proc is None:
            max_proc = proc

        dproc = super(DedicatedProcessorConfiguration, cls)._bld(adapter)

        dproc.desired = proc
        dproc.max = max_proc
        dproc.min = min_proc

        return dproc

    @property
    def desired(self):
        return self._get_val_int(_DPC_DES_PROCS, 0)

    @desired.setter
    def desired(self, value):
        self.set_parm_value(_DPC_DES_PROCS, value)

    @property
    def max(self):
        return self._get_val_int(_DPC_MAX_PROCS, 0)

    @max.setter
    def max(self, value):
        self.set_parm_value(_DPC_MAX_PROCS, value)

    @property
    def min(self):
        return self._get_val_int(_DPC_MIN_PROCS, 0)

    @min.setter
    def min(self, value):
        self.set_parm_value(_DPC_MIN_PROCS, value)


@ewrap.ElementWrapper.pvm_type('PartitionIOConfiguration', has_metadata=True)
class PartitionIOConfiguration(ewrap.ElementWrapper):
    """Represents the partitions Dedicated IO Configuration.

    Comprised of I/O Slots.  There are two types of IO slots.  Those dedicated
    to physical hardware (io_slots) and those that get used by virtual
    hardware.
    """

    @classmethod
    def bld(cls, adapter, max_virt_slots):
        """Builds a Partition IO configuration wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param max_virt_slots: Number of virtual slots (int)
        :returns: Partition IO configuration wrapper

        """
        cfg = super(PartitionIOConfiguration, cls)._bld(adapter)
        cfg.max_virtual_slots = max_virt_slots

        return cfg

    @property
    def max_virtual_slots(self):
        """The maximum number of virtual slots.

        Slots are used for every VirtualScsiServerAdapter, TrunkAdapter, etc...
        """
        return self._get_val_int(_IO_MAX_SLOTS)

    @max_virtual_slots.setter
    def max_virtual_slots(self, value):
        self.set_parm_value(_IO_MAX_SLOTS, value)

    @property
    def io_slots(self):
        """The physical I/O Slots.

        Each slot will have hardware associated with it.
        """
        es = ewrap.WrapperElemList(self._find_or_seed(IO_SLOTS_ROOT), IOSlot)
        return es

    @io_slots.setter
    def io_slots(self, val):
        self.replace_list(IO_SLOTS_ROOT, val)

    @property
    def tagged_io(self):
        """IBMi only - tagged I/O attributes of the I/O configuration."""
        tio = self._find(_IO_TIO)
        return TaggedIO.wrap(tio) if tio else None

    @tagged_io.setter
    def tagged_io(self, tio):
        self.inject(tio.element)


@ewrap.ElementWrapper.pvm_type('TaggedIO', has_metadata=True,
                               child_order=_TIO_EL_ORDER)
class TaggedIO(ewrap.ElementWrapper):
    """IBMi only - tagged I/O attributes of the I/O configuration."""

    @classmethod
    def bld(cls, adapter, load_src='0', console='HMC', alt_load_src='NONE'):
        """Builds a Partition TaggedIO wrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param load_src: Load source to use
        :param console: Console to use for IBMi
        :param alt_load_src: Alternate load source to use
        :returns: Partition TaggedIO wrapper

        """
        cfg = super(TaggedIO, cls)._bld(adapter)
        cfg.load_src = load_src
        cfg.console = console
        cfg.alt_load_src = alt_load_src
        return cfg

    @property
    def alt_load_src(self):
        """Value may or may not be an integer - always returned as string."""
        return self._get_val_str(_TIO_ALT_LOAD_SRC)

    @alt_load_src.setter
    def alt_load_src(self, value):
        self.set_parm_value(_TIO_ALT_LOAD_SRC, value)

    @property
    def console(self):
        """Value may or may not be an integer - always returned as string."""
        return self._get_val_str(_TIO_CONSOLE)

    @console.setter
    def console(self, value):
        self.set_parm_value(_TIO_CONSOLE, value)

    @property
    def load_src(self):
        """Value may or may not be an integer - always returned as string."""
        return self._get_val_str(_TIO_LOAD_SRC)

    @load_src.setter
    def load_src(self, value):
        self.set_parm_value(_TIO_LOAD_SRC, value)


@ewrap.ElementWrapper.pvm_type('ProfileIOSlot', has_metadata=True,
                               child_order=_IO_SLOT_ORDER)
class IOSlot(ewrap.ElementWrapper):
    """An I/O Slot represents a device bus on the system.

    It may contain a piece of hardware within it.
    """
    @classmethod
    def bld(cls, adapter, bus_grp_required, drc_index, required=False):
        """Build a new IOSlot wrapper with all required parameters.

        :returns: A new IOSlot wrapper.
        """
        new_slot = super(IOSlot, cls)._bld(adapter)

        new_slot.required = required

        # Build out the AssociatedIOSlot
        assoc_io_slot = cls.AssociatedIOSlot._bld_new(adapter,
                                                      bus_grp_required,
                                                      drc_index)

        # Inject the AssociatedIOSlot into this wrapper
        new_slot.inject(assoc_io_slot.element)

        return new_slot

    @ewrap.ElementWrapper.pvm_type(ASSOC_IO_SLOT_ROOT, has_metadata=True,
                                   child_order=_AIO_ORDER)
    class AssociatedIOSlot(ewrap.ElementWrapper):
        """Internal class.  Hides the nested AssociatedIOSlot from parent.

        Every ProfileIOSlot contains one AssociatedIOSlot.  If both are
        exposed at the API level, the user would have to go from:
         - lpar -> partition i/o config -> i/o slot -> associated i/o slot ->
           i/o data

        Since every i/o slot has a single Associated I/O Slot (unless said
        I/O slot has no associated I/O), then we can just hide this from
        the user.

        We still keep the structure internally, but makes the API easier to
        consume.
        """
        @classmethod
        def _bld_new(cls, adapter, bus_grp_required, drc_index):
            """Build a new AssociatedIOSlot wrapper.

            This will not typically be called outside of the IOSlot.bld
            class method.

            :returns: A new AssociatedIOSlot wrapper.
            """
            new_slot = super(IOSlot.AssociatedIOSlot, cls)._bld(adapter)
            new_slot._bus_grp_required(bus_grp_required)
            new_slot._drc_index(drc_index)

            return new_slot

        @property
        def bus_grp_required(self):
            return self._get_val_bool(_ASSOC_IO_SLOT_BUS_GRP)

        def _bus_grp_required(self, val):
            self.set_parm_value(_ASSOC_IO_SLOT_BUS_GRP,
                                u.sanitize_bool_for_api(val))

        @property
        def description(self):
            return self._get_val_str(_ASSOC_IO_SLOT_DESC)

        @property
        def phys_loc(self):
            return self._get_val_str(_ASSOC_IO_SLOT_PHYS_LOC)

        @property
        def pc_adpt_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_ADPT_ID)

        @property
        def pci_class(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_CLASS)

        @property
        def pci_dev_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_DEV_ID)

        @property
        def pci_subsys_dev_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_SUBSYS_DEV_ID)

        @property
        def pci_mfg_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_MFG_ID)

        @property
        def pci_rev_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_REV_ID)

        @property
        def pci_vendor_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_PCI_VENDOR_ID)

        @property
        def pci_subsys_vendor_id(self):
            return self._get_val_int(_ASSOC_IO_SLOT_SUBSYS_VENDOR_ID)

        @property
        def drc_index(self):
            return self._get_val_int(_ASSOC_IO_SLOT_DRC_INDEX)

        def _drc_index(self, val):
            self.set_parm_value(_ASSOC_IO_SLOT_DRC_INDEX, val)

        @property
        def drc_name(self):
            return self._get_val_str(_ASSOC_IO_SLOT_DRC_NAME)

        @property
        def io_adapter(self):
            """Jumps over the 'Related IO Adapter' element direct to the I/O.

            This is another area where the schema has a two step jump that the
            API can avoid.  This method skips over the RelatedIOAdapter
            and jumps right to the IO Adapter.

            Return values are either the generic IOAdapter or the
            PhysFCAdapter.
            """
            # The child can be either an IO Adapter or a PhysFCAdapter.
            # Need to check for both...
            io_adpt_root = self._find(
                u.xpath(RELATED_IO_ADPT_ROOT, IOAdapter.schema_type))
            if io_adpt_root is not None:
                return IOAdapter.wrap(io_adpt_root)

            # Didn't have the generic...check for non-generic.
            io_adpt_root = self._find(
                u.xpath(RELATED_IO_ADPT_ROOT, PhysFCAdapter.schema_type))
            if io_adpt_root is not None:
                return PhysFCAdapter.wrap(io_adpt_root)

            return None

    @property
    def required(self):
        return self._get_val_bool(_IO_SLOT_REQ)

    @required.setter
    def required(self, val):
        self.set_parm_value(_IO_SLOT_REQ, u.sanitize_bool_for_api(val))

    def __get_prop(self, func):
        """Thin wrapper to get the Associated I/O Slot and get a property."""
        elem = self._find(ASSOC_IO_SLOT_ROOT)
        if elem is None:
            return None

        # Build the Associated IO Slot, find the function and execute it.
        assoc_io_slot = self.AssociatedIOSlot.wrap(elem)
        return getattr(assoc_io_slot, func)

    @property
    def bus_grp_required(self):
        return self.__get_prop('bus_grp_required')

    @property
    def description(self):
        return self.__get_prop('description')

    @property
    def phys_loc(self):
        return self.__get_prop('phys_loc')

    @property
    def pc_adpt_id(self):
        return self.__get_prop('pc_adpt_id')

    @property
    def pci_class(self):
        return self.__get_prop('pci_class')

    @property
    def pci_dev_id(self):
        return self.__get_prop('pci_dev_id')

    @property
    def pci_subsys_dev_id(self):
        return self.__get_prop('pci_subsys_dev_id')

    @property
    def pci_mfg_id(self):
        return self.__get_prop('pci_mfg_id')

    @property
    def pci_rev_id(self):
        return self.__get_prop('pci_rev_id')

    @property
    def pci_vendor_id(self):
        return self.__get_prop('pci_vendor_id')

    @property
    def pci_subsys_vendor_id(self):
        return self.__get_prop('pci_subsys_vendor_id')

    @property
    def drc_index(self):
        return self.__get_prop('drc_index')

    @property
    def drc_name(self):
        return self.__get_prop('drc_name')

    @property
    def adapter(self):
        """DEPRECATED - use 'io_adapter' method instead."""
        import warnings
        warnings.warn(
            _("IOSlot.adapter is deprecated!  Use IOSlot.io_adapter instead."),
            DeprecationWarning)
        return self.io_adapter

    @property
    def io_adapter(self):
        """Returns the physical I/O Adapter for this slot.

        This will be one of two types.  Either a generic I/O Adapter or
        a Physical Fibre Channel Adapter (PhysFCAdapter).
        """
        return self.__get_prop('io_adapter')
